# maybe rename this `main.py` or `venom.py`
# (can have an `__init__.py` which exposes the API).

from typing import Any, Optional

from vyper.codegen.ir_node import IRnode
from vyper.compiler.settings import OptimizationLevel
from vyper.venom.analysis import DFG, calculate_cfg, calculate_liveness
from vyper.venom.bb_optimizer import (
    ir_pass_optimize_empty_blocks,
    ir_pass_optimize_unused_variables,
    ir_pass_remove_unreachable_blocks,
)
from vyper.venom.function import IRFunction
from vyper.venom.ir_node_to_venom import ir_node_to_venom
from vyper.venom.passes.constant_propagation import ir_pass_constant_propagation
from vyper.venom.passes.dft import DFTPass
from vyper.venom.venom_to_assembly import VenomCompiler

DEFAULT_OPT_LEVEL = OptimizationLevel.default()


def generate_assembly_experimental(
    runtime_code: IRFunction,
    deploy_code: Optional[IRFunction] = None,
    optimize: OptimizationLevel = DEFAULT_OPT_LEVEL,
) -> list[str]:
    # note: VenomCompiler is sensitive to the order of these!
    if deploy_code is not None:
        functions = [deploy_code, runtime_code]
    else:
        functions = [runtime_code]

    compiler = VenomCompiler(functions)
    return compiler.generate_evm(optimize == OptimizationLevel.NONE)


def _run_passes(ctx: IRFunction, optimize: OptimizationLevel) -> None:
    # Run passes on Venom IR
    # TODO: Add support for optimization levels
    while True:
        changes = 0

        changes += ir_pass_optimize_empty_blocks(ctx)
        changes += ir_pass_remove_unreachable_blocks(ctx)

        calculate_liveness(ctx)

        changes += ir_pass_optimize_unused_variables(ctx)

        calculate_cfg(ctx)
        calculate_liveness(ctx)

        changes += ir_pass_constant_propagation(ctx)
        changes += DFTPass.run_pass(ctx)

        calculate_cfg(ctx)
        calculate_liveness(ctx)

        if changes == 0:
            break


def generate_ir(ir: IRnode, optimize: OptimizationLevel) -> IRFunction:
    # Convert "old" IR to "new" IR
    ctx = ir_node_to_venom(ir)
    _run_passes(ctx, optimize)

    return ctx



import itertools

import pytest


def test_bytes_literal_code(get_contract_with_gas_estimation):
    bytes_literal_code = """
@external
def foo() -> Bytes[5]:
    return b"horse"

@external
def bar() -> Bytes[10]:
    return concat(b"b", b"a", b"d", b"m", b"i", b"", b"nton")

@external
def baz() -> Bytes[40]:
    return concat(b"0123456789012345678901234567890", b"12")

@external
def baz2() -> Bytes[40]:
    return concat(b"01234567890123456789012345678901", b"12")

@external
def baz3() -> Bytes[40]:
    return concat(b"0123456789012345678901234567890", b"1")

@external
def baz4() -> Bytes[100]:
    return concat(b"01234567890123456789012345678901234567890123456789",
                  b"01234567890123456789012345678901234567890123456789")
    """

    c = get_contract_with_gas_estimation(bytes_literal_code)
    assert c.foo() == b"horse"
    assert c.bar() == b"badminton"
    assert c.baz() == b"012345678901234567890123456789012"
    assert c.baz2() == b"0123456789012345678901234567890112"
    assert c.baz3() == b"01234567890123456789012345678901"
    assert c.baz4() == b"0123456789" * 10

    print("Passed string literal test")


@pytest.mark.parametrize("i,e,_s", itertools.product([95, 96, 97], [63, 64, 65], [31, 32, 33]))
def test_bytes_literal_splicing_fuzz(get_contract_with_gas_estimation, i, e, _s):
    kode = f"""
moo: Bytes[100]

@external
def foo(s: uint256, L: uint256) -> Bytes[100]:
    x: int128 = 27
    r: Bytes[100] = slice(b"{("c" * i)}", s, L)
    y: int128 = 37
    if x * y == 999:
        return r
    return b"3434346667777"

@external
def bar(s: uint256, L: uint256) -> Bytes[100]:
    self.moo = b"{("c" * i)}"
    x: int128 = 27
    r: Bytes[100] = slice(self.moo, s, L)
    y: int128  = 37
    if x * y == 999:
        return r
    return b"3434346667777"

@external
def baz(s: uint256, L: uint256) -> Bytes[100]:
    x: int128 = 27
    self.moo = slice(b"{("c" * i)}", s, L)
    y: int128 = 37
    if x * y == 999:
        return self.moo
    return b"3434346667777"
    """

    c = get_contract_with_gas_estimation(kode)
    o1 = c.foo(_s, e - _s)
    o2 = c.bar(_s, e - _s)
    o3 = c.baz(_s, e - _s)
    assert o1 == o2 == o3 == b"c" * (e - _s), (i, _s, e - _s, o1, o2, o3)

    print("Passed string literal splicing fuzz-test")



# TODO: check, this is probably redundant with examples/test_crowdfund.py
def test_crowdfund(w3, tester, get_contract_with_gas_estimation_for_constants):
    crowdfund = """

struct Funder:
    sender: address
    value: uint256
funders: HashMap[int128, Funder]
nextFunderIndex: int128
beneficiary: address
deadline: public(uint256)
goal: public(uint256)
refundIndex: int128
timelimit: public(uint256)

@deploy
def __init__(_beneficiary: address, _goal: uint256, _timelimit: uint256):
    self.beneficiary = _beneficiary
    self.deadline = block.timestamp + _timelimit
    self.timelimit = _timelimit
    self.goal = _goal

@external
@payable
def participate():
    assert block.timestamp < self.deadline
    nfi: int128 = self.nextFunderIndex
    self.funders[nfi].sender = msg.sender
    self.funders[nfi].value = msg.value
    self.nextFunderIndex = nfi + 1

@external
@view
def expired() -> bool:
    return block.timestamp >= self.deadline

@external
@view
def block_timestamp() -> uint256:
    return block.timestamp

@external
@view
def reached() -> bool:
    return self.balance >= self.goal

@external
def finalize():
    assert block.timestamp >= self.deadline and self.balance >= self.goal
    selfdestruct(self.beneficiary)

@external
def refund():
    ind: int128 = self.refundIndex
    for i: int128 in range(ind, ind + 30, bound=30):
        if i >= self.nextFunderIndex:
            self.refundIndex = self.nextFunderIndex
            return
        send(self.funders[i].sender, self.funders[i].value)
        self.funders[i].sender = 0x0000000000000000000000000000000000000000
        self.funders[i].value = 0
    self.refundIndex = ind + 30

    """
    a0, a1, a2, a3, a4, a5, a6 = w3.eth.accounts[:7]

    c = get_contract_with_gas_estimation_for_constants(crowdfund, *[a1, 50, 60])
    start_timestamp = w3.eth.get_block(w3.eth.block_number).timestamp

    c.participate(transact={"value": 5})
    assert c.timelimit() == 60
    assert c.deadline() - start_timestamp == 60
    assert not c.expired()
    assert not c.reached()
    c.participate(transact={"value": 49})
    assert c.reached()
    pre_bal = w3.eth.get_balance(a1)
    w3.testing.mine(100)
    assert c.expired()
    c.finalize(transact={})
    post_bal = w3.eth.get_balance(a1)
    assert post_bal - pre_bal == 54

    c = get_contract_with_gas_estimation_for_constants(crowdfund, *[a1, 50, 60])
    c.participate(transact={"value": 1, "from": a3})
    c.participate(transact={"value": 2, "from": a4})
    c.participate(transact={"value": 3, "from": a5})
    c.participate(transact={"value": 4, "from": a6})
    w3.testing.mine(100)
    assert c.expired()
    assert not c.reached()
    pre_bals = [w3.eth.get_balance(x) for x in [a3, a4, a5, a6]]
    c.refund(transact={})
    post_bals = [w3.eth.get_balance(x) for x in [a3, a4, a5, a6]]
    assert [y - x for x, y in zip(pre_bals, post_bals)] == [1, 2, 3, 4]


def test_crowdfund2(w3, tester, get_contract_with_gas_estimation_for_constants):
    crowdfund2 = """
struct Funder:
    sender: address
    value: uint256

funders: HashMap[int128, Funder]
nextFunderIndex: int128
beneficiary: address
deadline: public(uint256)
goal: uint256
refundIndex: int128
timelimit: public(uint256)

@deploy
def __init__(_beneficiary: address, _goal: uint256, _timelimit: uint256):
    self.beneficiary = _beneficiary
    self.deadline = block.timestamp + _timelimit
    self.timelimit = _timelimit
    self.goal = _goal

@external
@payable
def participate():
    assert block.timestamp < self.deadline
    nfi: int128 = self.nextFunderIndex
    self.funders[nfi] = Funder(sender=msg.sender, value=msg.value)
    self.nextFunderIndex = nfi + 1

@external
@view
def expired() -> bool:
    return block.timestamp >= self.deadline

@external
@view
def block_timestamp() -> uint256:
    return block.timestamp

@external
@view
def reached() -> bool:
    return self.balance >= self.goal

@external
def finalize():
    assert block.timestamp >= self.deadline and self.balance >= self.goal
    selfdestruct(self.beneficiary)

@external
def refund():
    ind: int128 = self.refundIndex
    for i: int128 in range(ind, ind + 30, bound=30):
        if i >= self.nextFunderIndex:
            self.refundIndex = self.nextFunderIndex
            return
        send(self.funders[i].sender, self.funders[i].value)
        self.funders[i] = empty(Funder)
    self.refundIndex = ind + 30

    """
    a0, a1, a2, a3, a4, a5, a6 = w3.eth.accounts[:7]
    c = get_contract_with_gas_estimation_for_constants(crowdfund2, *[a1, 50, 60])

    c.participate(transact={"value": 5})
    assert c.timelimit() == 60
    assert c.deadline() - c.block_timestamp() == 59
    assert not c.expired()
    assert not c.reached()
    c.participate(transact={"value": 49})
    assert c.reached()
    pre_bal = w3.eth.get_balance(a1)
    w3.testing.mine(100)
    assert c.expired()
    c.finalize(transact={})
    post_bal = w3.eth.get_balance(a1)
    assert post_bal - pre_bal == 54

    c = get_contract_with_gas_estimation_for_constants(crowdfund2, *[a1, 50, 60])
    c.participate(transact={"value": 1, "from": a3})
    c.participate(transact={"value": 2, "from": a4})
    c.participate(transact={"value": 3, "from": a5})
    c.participate(transact={"value": 4, "from": a6})
    w3.testing.mine(100)
    assert c.expired()
    assert not c.reached()
    pre_bals = [w3.eth.get_balance(x) for x in [a3, a4, a5, a6]]
    c.refund(transact={})
    post_bals = [w3.eth.get_balance(x) for x in [a3, a4, a5, a6]]
    assert [y - x for x, y in zip(pre_bals, post_bals)] == [1, 2, 3, 4]



import json
import logging
from contextlib import contextmanager
from functools import wraps

import hypothesis
import pytest
import web3.exceptions
from eth_tester import EthereumTester, PyEVMBackend
from eth_tester.exceptions import TransactionFailed
from eth_utils import setup_DEBUG2_logging
from eth_utils.toolz import compose
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract
from web3.providers.eth_tester import EthereumTesterProvider

from tests.utils import working_directory
from vyper import compiler
from vyper.ast.grammar import parse_vyper_source
from vyper.codegen.ir_node import IRnode
from vyper.compiler.input_bundle import FilesystemInputBundle, InputBundle
from vyper.compiler.settings import OptimizationLevel, Settings, _set_debug_mode
from vyper.ir import compile_ir, optimizer
from vyper.utils import ERC5202_PREFIX

# Import the base fixtures
pytest_plugins = ["tests.fixtures.memorymock"]

############
# PATCHING #
############


# disable hypothesis deadline globally
hypothesis.settings.register_profile("ci", deadline=None)
hypothesis.settings.load_profile("ci")


def set_evm_verbose_logging():
    logger = logging.getLogger("eth.vm.computation.Computation")
    setup_DEBUG2_logging()
    logger.setLevel("DEBUG2")


# Useful options to comment out whilst working:
# set_evm_verbose_logging()
#
# from vdb import vdb
# vdb.set_evm_opcode_debugger()


def pytest_addoption(parser):
    parser.addoption(
        "--optimize",
        choices=["codesize", "gas", "none"],
        default="gas",
        help="change optimization mode",
    )
    parser.addoption("--enable-compiler-debug-mode", action="store_true")


@pytest.fixture(scope="module")
def output_formats():
    output_formats = compiler.OUTPUT_FORMATS.copy()
    del output_formats["bb"]
    del output_formats["bb_runtime"]
    return output_formats


@pytest.fixture(scope="module")
def optimize(pytestconfig):
    flag = pytestconfig.getoption("optimize")
    return OptimizationLevel.from_string(flag)


@pytest.fixture(scope="session", autouse=True)
def debug(pytestconfig):
    debug = pytestconfig.getoption("enable_compiler_debug_mode")
    assert isinstance(debug, bool)
    _set_debug_mode(debug)


@pytest.fixture
def chdir_tmp_path(tmp_path):
    # this is useful for when you want imports to have relpaths
    with working_directory(tmp_path):
        yield


@pytest.fixture
def keccak():
    return Web3.keccak


@pytest.fixture
def make_file(tmp_path):
    # writes file_contents to file_name, creating it in the
    # tmp_path directory. returns final path.
    def fn(file_name, file_contents):
        path = tmp_path / file_name
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w") as f:
            f.write(file_contents)

        return path

    return fn


# this can either be used for its side effects (to prepare a call
# to get_contract), or the result can be provided directly to
# compile_code / CompilerData.
@pytest.fixture
def make_input_bundle(tmp_path, make_file):
    def fn(sources_dict):
        for file_name, file_contents in sources_dict.items():
            make_file(file_name, file_contents)
        return FilesystemInputBundle([tmp_path])

    return fn


# for tests which just need an input bundle, doesn't matter what it is
@pytest.fixture
def dummy_input_bundle():
    return InputBundle([])


# TODO: remove me, this is just string.encode("utf-8").ljust()
# only used in test_logging.py.
@pytest.fixture
def bytes_helper():
    def bytes_helper(str, length):
        return bytes(str, "utf-8") + bytearray(length - len(str))

    return bytes_helper


def _none_addr(datatype, data):
    if datatype == "address" and int(data, base=16) == 0:
        return (datatype, None)
    else:
        return (datatype, data)


CONCISE_NORMALIZERS = (_none_addr,)


@pytest.fixture(scope="module")
def tester():
    # set absurdly high gas limit so that london basefee never adjusts
    # (note: 2**63 - 1 is max that evm allows)
    custom_genesis = PyEVMBackend._generate_genesis_params(overrides={"gas_limit": 10**10})
    custom_genesis["base_fee_per_gas"] = 0
    backend = PyEVMBackend(genesis_parameters=custom_genesis)
    return EthereumTester(backend=backend)


def zero_gas_price_strategy(web3, transaction_params=None):
    return 0  # zero gas price makes testing simpler.


@pytest.fixture(scope="module")
def w3(tester):
    w3 = Web3(EthereumTesterProvider(tester))
    w3.eth.set_gas_price_strategy(zero_gas_price_strategy)
    return w3


def get_compiler_gas_estimate(code, func):
    sigs = compiler.phases.CompilerData(code).function_signatures
    if func:
        return compiler.utils.build_gas_estimates(sigs)[func] + 22000
    else:
        return sum(compiler.utils.build_gas_estimates(sigs).values()) + 22000


def check_gas_on_chain(w3, tester, code, func=None, res=None):
    gas_estimate = get_compiler_gas_estimate(code, func)
    gas_actual = tester.get_block_by_number("latest")["gas_used"]
    # Computed upper bound on the gas consumption should
    # be greater than or equal to the amount of gas used
    if gas_estimate < gas_actual:
        raise Exception(f"Gas upper bound fail: bound {gas_estimate} actual {gas_actual}")

    print(f"Function name: {func} - Gas estimate {gas_estimate}, Actual: {gas_actual}")


def gas_estimation_decorator(w3, tester, fn, source_code, func):
    def decorator(*args, **kwargs):
        @wraps(fn)
        def decorated_function(*args, **kwargs):
            result = fn(*args, **kwargs)
            if "transact" in kwargs:
                check_gas_on_chain(w3, tester, source_code, func, res=result)
            return result

        return decorated_function(*args, **kwargs)

    return decorator


def set_decorator_to_contract_function(w3, tester, contract, source_code, func):
    func_definition = getattr(contract, func)
    func_with_decorator = gas_estimation_decorator(w3, tester, func_definition, source_code, func)
    setattr(contract, func, func_with_decorator)


class VyperMethod:
    ALLOWED_MODIFIERS = {"call", "estimateGas", "transact", "buildTransaction"}

    def __init__(self, function, normalizers=None):
        self._function = function
        self._function._return_data_normalizers = normalizers

    def __call__(self, *args, **kwargs):
        return self.__prepared_function(*args, **kwargs)

    def __prepared_function(self, *args, **kwargs):
        if not kwargs:
            modifier, modifier_dict = "call", {}
            fn_abi = [
                x
                for x in self._function.contract_abi
                if x.get("name") == self._function.function_identifier
            ].pop()
            # To make tests faster just supply some high gas value.
            modifier_dict.update({"gas": fn_abi.get("gas", 0) + 500000})
        elif len(kwargs) == 1:
            modifier, modifier_dict = kwargs.popitem()
            if modifier not in self.ALLOWED_MODIFIERS:
                raise TypeError(f"The only allowed keyword arguments are: {self.ALLOWED_MODIFIERS}")
        else:
            raise TypeError(f"Use up to one keyword argument, one of: {self.ALLOWED_MODIFIERS}")
        return getattr(self._function(*args), modifier)(modifier_dict)


class VyperContract:
    """
    An alternative Contract Factory which invokes all methods as `call()`,
    unless you add a keyword argument. The keyword argument assigns the prep method.
    This call
    > contract.withdraw(amount, transact={'from': eth.accounts[1], 'gas': 100000, ...})
    is equivalent to this call in the classic contract:
    > contract.functions.withdraw(amount).transact({'from': eth.accounts[1], 'gas': 100000, ...})
    """

    def __init__(self, classic_contract, method_class=VyperMethod):
        classic_contract._return_data_normalizers += CONCISE_NORMALIZERS
        self._classic_contract = classic_contract
        self.address = self._classic_contract.address
        protected_fn_names = [fn for fn in dir(self) if not fn.endswith("__")]

        try:
            fn_names = [fn["name"] for fn in self._classic_contract.functions._functions]
        except web3.exceptions.NoABIFunctionsFound:
            fn_names = []

        for fn_name in fn_names:
            # Override namespace collisions
            if fn_name in protected_fn_names:
                raise AttributeError(f"{fn_name} is protected!")
            else:
                _classic_method = getattr(self._classic_contract.functions, fn_name)
                _concise_method = method_class(
                    _classic_method, self._classic_contract._return_data_normalizers
                )
            setattr(self, fn_name, _concise_method)

    @classmethod
    def factory(cls, *args, **kwargs):
        return compose(cls, Contract.factory(*args, **kwargs))


@pytest.fixture
def get_contract_from_ir(w3, optimize):
    def ir_compiler(ir, *args, **kwargs):
        ir = IRnode.from_list(ir)
        if optimize != OptimizationLevel.NONE:
            ir = optimizer.optimize(ir)

        bytecode, _ = compile_ir.assembly_to_evm(
            compile_ir.compile_to_assembly(ir, optimize=optimize)
        )

        abi = kwargs.get("abi") or []
        c = w3.eth.contract(abi=abi, bytecode=bytecode)
        deploy_transaction = c.constructor()
        tx_hash = deploy_transaction.transact()
        address = w3.eth.get_transaction_receipt(tx_hash)["contractAddress"]
        contract = w3.eth.contract(
            address, abi=abi, bytecode=bytecode, ContractFactoryClass=VyperContract
        )
        return contract

    return ir_compiler


def _get_contract(
    w3,
    source_code,
    optimize,
    output_formats,
    *args,
    override_opt_level=None,
    input_bundle=None,
    **kwargs,
):
    settings = Settings()
    settings.evm_version = kwargs.pop("evm_version", None)
    settings.optimize = override_opt_level or optimize
    out = compiler.compile_code(
        source_code,
        # test that all output formats can get generated
        output_formats=output_formats,
        settings=settings,
        input_bundle=input_bundle,
        show_gas_estimates=True,  # Enable gas estimates for testing
    )
    parse_vyper_source(source_code)  # Test grammar.
    json.dumps(out["metadata"])  # test metadata is json serializable
    abi = out["abi"]
    bytecode = out["bytecode"]
    value = kwargs.pop("value_in_eth", 0) * 10**18  # Handle deploying with an eth value.
    c = w3.eth.contract(abi=abi, bytecode=bytecode)
    deploy_transaction = c.constructor(*args)
    tx_info = {"from": w3.eth.accounts[0], "value": value, "gasPrice": 0}
    tx_info.update(kwargs)
    tx_hash = deploy_transaction.transact(tx_info)
    address = w3.eth.get_transaction_receipt(tx_hash)["contractAddress"]
    return w3.eth.contract(address, abi=abi, bytecode=bytecode, ContractFactoryClass=VyperContract)


@pytest.fixture(scope="module")
def get_contract(w3, optimize, output_formats):
    def fn(source_code, *args, **kwargs):
        return _get_contract(w3, source_code, optimize, output_formats, *args, **kwargs)

    return fn


@pytest.fixture
def get_contract_with_gas_estimation(tester, w3, optimize, output_formats):
    def get_contract_with_gas_estimation(source_code, *args, **kwargs):
        contract = _get_contract(w3, source_code, optimize, output_formats, *args, **kwargs)
        for abi_ in contract._classic_contract.functions.abi:
            if abi_["type"] == "function":
                set_decorator_to_contract_function(w3, tester, contract, source_code, abi_["name"])
        return contract

    return get_contract_with_gas_estimation


@pytest.fixture
def get_contract_with_gas_estimation_for_constants(w3, optimize, output_formats):
    def get_contract_with_gas_estimation_for_constants(source_code, *args, **kwargs):
        return _get_contract(w3, source_code, optimize, output_formats, *args, **kwargs)

    return get_contract_with_gas_estimation_for_constants


@pytest.fixture(scope="module")
def get_contract_module(optimize, output_formats):
    """
    This fixture is used for Hypothesis tests to ensure that
    the same contract is called over multiple runs of the test.
    """
    custom_genesis = PyEVMBackend._generate_genesis_params(overrides={"gas_limit": 4500000})
    custom_genesis["base_fee_per_gas"] = 0
    backend = PyEVMBackend(genesis_parameters=custom_genesis)
    tester = EthereumTester(backend=backend)
    w3 = Web3(EthereumTesterProvider(tester))
    w3.eth.set_gas_price_strategy(zero_gas_price_strategy)

    def get_contract_module(source_code, *args, **kwargs):
        return _get_contract(w3, source_code, optimize, output_formats, *args, **kwargs)

    return get_contract_module


def _deploy_blueprint_for(
    w3, source_code, optimize, output_formats, initcode_prefix=ERC5202_PREFIX, **kwargs
):
    settings = Settings()
    settings.evm_version = kwargs.pop("evm_version", None)
    settings.optimize = optimize
    out = compiler.compile_code(
        source_code,
        output_formats=output_formats,
        settings=settings,
        show_gas_estimates=True,  # Enable gas estimates for testing
    )
    parse_vyper_source(source_code)  # Test grammar.
    abi = out["abi"]
    bytecode = HexBytes(initcode_prefix) + HexBytes(out["bytecode"])
    bytecode_len = len(bytecode)
    bytecode_len_hex = hex(bytecode_len)[2:].rjust(4, "0")
    # prepend a quick deploy preamble
    deploy_preamble = HexBytes("61" + bytecode_len_hex + "3d81600a3d39f3")
    deploy_bytecode = HexBytes(deploy_preamble) + bytecode

    deployer_abi = []  # just a constructor
    c = w3.eth.contract(abi=deployer_abi, bytecode=deploy_bytecode)
    deploy_transaction = c.constructor()
    tx_info = {"from": w3.eth.accounts[0], "value": 0, "gasPrice": 0}

    tx_hash = deploy_transaction.transact(tx_info)
    address = w3.eth.get_transaction_receipt(tx_hash)["contractAddress"]

    # sanity check
    assert w3.eth.get_code(address) == bytecode, (w3.eth.get_code(address), bytecode)

    def factory(address):
        return w3.eth.contract(
            address, abi=abi, bytecode=bytecode, ContractFactoryClass=VyperContract
        )

    return w3.eth.contract(address, bytecode=deploy_bytecode), factory


@pytest.fixture(scope="module")
def deploy_blueprint_for(w3, optimize, output_formats):
    def deploy_blueprint_for(source_code, *args, **kwargs):
        return _deploy_blueprint_for(w3, source_code, optimize, output_formats, *args, **kwargs)

    return deploy_blueprint_for


# TODO: this should not be a fixture.
# remove me and replace all uses with `with pytest.raises`.
@pytest.fixture
def assert_compile_failed():
    def assert_compile_failed(function_to_test, exception=Exception):
        with pytest.raises(exception):
            function_to_test()

    return assert_compile_failed


@pytest.fixture
def create2_address_of(keccak):
    def _f(_addr, _salt, _initcode):
        prefix = HexBytes("0xff")
        addr = HexBytes(_addr)
        salt = HexBytes(_salt)
        initcode = HexBytes(_initcode)
        return keccak(prefix + addr + salt + keccak(initcode))[12:]

    return _f


@pytest.fixture
def side_effects_contract(get_contract):
    def generate(ret_type):
        """
        Generates a Vyper contract with an external `foo()` function, which
        returns the specified return value of the specified return type, for
        testing side effects using the `assert_side_effects_invoked` fixture.
        """
        code = f"""
counter: public(uint256)

@external
def foo(s: {ret_type}) -> {ret_type}:
    self.counter += 1
    return s
    """
        contract = get_contract(code)
        return contract

    return generate


@pytest.fixture
def assert_side_effects_invoked():
    def assert_side_effects_invoked(side_effects_contract, side_effects_trigger, n=1):
        start_value = side_effects_contract.counter()

        side_effects_trigger()

        end_value = side_effects_contract.counter()
        assert end_value == start_value + n

    return assert_side_effects_invoked


@pytest.fixture
def get_logs(w3):
    def get_logs(tx_hash, c, event_name):
        tx_receipt = w3.eth.get_transaction_receipt(tx_hash)
        return c._classic_contract.events[event_name]().process_receipt(tx_receipt)

    return get_logs


@pytest.fixture(scope="module")
def tx_failed(tester):
    @contextmanager
    def fn(exception=TransactionFailed, exc_text=None):
        snapshot_id = tester.take_snapshot()
        with pytest.raises(exception) as excinfo:
            yield excinfo
        tester.revert_to_snapshot(snapshot_id)
        if exc_text:
            # TODO test equality
            assert exc_text in str(excinfo.value), (exc_text, excinfo.value)

    return fn



from typing import Optional

from vyper.venom.basicblock import (
    IRBasicBlock,
    IRInstruction,
    IRLabel,
    IROperand,
    IRVariable,
    MemType,
)

GLOBAL_LABEL = IRLabel("__global")


class IRFunction:
    """
    Function that contains basic blocks.
    """

    name: IRLabel  # symbol name
    entry_points: list[IRLabel]  # entry points
    args: list
    ctor_mem_size: Optional[int]
    immutables_len: Optional[int]
    basic_blocks: list[IRBasicBlock]
    data_segment: list[IRInstruction]
    last_label: int
    last_variable: int

    def __init__(self, name: IRLabel = None) -> None:
        if name is None:
            name = GLOBAL_LABEL
        self.name = name
        self.entry_points = []
        self.args = []
        self.ctor_mem_size = None
        self.immutables_len = None
        self.basic_blocks = []
        self.data_segment = []
        self.last_label = 0
        self.last_variable = 0

        self.add_entry_point(name)
        self.append_basic_block(IRBasicBlock(name, self))

    def add_entry_point(self, label: IRLabel) -> None:
        """
        Add entry point.
        """
        self.entry_points.append(label)

    def remove_entry_point(self, label: IRLabel) -> None:
        """
        Remove entry point.
        """
        self.entry_points.remove(label)

    def append_basic_block(self, bb: IRBasicBlock) -> IRBasicBlock:
        """
        Append basic block to function.
        """
        assert isinstance(bb, IRBasicBlock), f"append_basic_block takes IRBasicBlock, got '{bb}'"
        self.basic_blocks.append(bb)

        # TODO add sanity check somewhere that basic blocks have unique labels

        return self.basic_blocks[-1]

    def get_basic_block(self, label: Optional[str] = None) -> IRBasicBlock:
        """
        Get basic block by label.
        If label is None, return the last basic block.
        """
        if label is None:
            return self.basic_blocks[-1]
        for bb in self.basic_blocks:
            if bb.label.value == label:
                return bb
        raise AssertionError(f"Basic block '{label}' not found")

    def get_basic_block_after(self, label: IRLabel) -> IRBasicBlock:
        """
        Get basic block after label.
        """
        for i, bb in enumerate(self.basic_blocks[:-1]):
            if bb.label.value == label.value:
                return self.basic_blocks[i + 1]
        raise AssertionError(f"Basic block after '{label}' not found")

    def get_basicblocks_in(self, basic_block: IRBasicBlock) -> list[IRBasicBlock]:
        """
        Get basic blocks that contain label.
        """
        return [bb for bb in self.basic_blocks if basic_block.label in bb.cfg_in]

    def get_next_label(self) -> IRLabel:
        self.last_label += 1
        return IRLabel(f"{self.last_label}")

    def get_next_variable(
        self, mem_type: MemType = MemType.OPERAND_STACK, mem_addr: Optional[int] = None
    ) -> IRVariable:
        self.last_variable += 1
        return IRVariable(f"%{self.last_variable}", mem_type, mem_addr)

    def get_last_variable(self) -> str:
        return f"%{self.last_variable}"

    def remove_unreachable_blocks(self) -> int:
        removed = 0
        new_basic_blocks = []
        for bb in self.basic_blocks:
            if not bb.is_reachable and bb.label not in self.entry_points:
                removed += 1
            else:
                new_basic_blocks.append(bb)
        self.basic_blocks = new_basic_blocks
        return removed

    def append_data(self, opcode: str, args: list[IROperand]) -> None:
        """
        Append data
        """
        self.data_segment.append(IRInstruction(opcode, args))  # type: ignore

    @property
    def normalized(self) -> bool:
        """
        Check if function is normalized. A function is normalized if in the
        CFG, no basic block simultaneously has multiple inputs and outputs.
        That is, a basic block can be jumped to *from* multiple blocks, or it
        can jump *to* multiple blocks, but it cannot simultaneously do both.
        Having a normalized CFG makes calculation of stack layout easier when
        emitting assembly.
        """
        for bb in self.basic_blocks:
            # Ignore if there are no multiple predecessors
            if len(bb.cfg_in) <= 1:
                continue

            # Check if there is a branching jump at the end
            # of one of the predecessors
            for in_bb in bb.cfg_in:
                if len(in_bb.cfg_out) > 1:
                    return False

        # The function is normalized
        return True

    def copy(self):
        new = IRFunction(self.name)
        new.basic_blocks = self.basic_blocks.copy()
        new.data_segment = self.data_segment.copy()
        new.last_label = self.last_label
        new.last_variable = self.last_variable
        return new

    def __repr__(self) -> str:
        str = f"IRFunction: {self.name}\n"
        for bb in self.basic_blocks:
            str += f"{bb}\n"
        if len(self.data_segment) > 0:
            str += "Data segment:\n"
            for inst in self.data_segment:
                str += f"{inst}\n"
        return str.strip()



import pytest
from web3.exceptions import ValidationError

INITIAL_VALUE = 4


@pytest.fixture
def adv_storage_contract(w3, get_contract):
    with open("examples/storage/advanced_storage.vy") as f:
        contract_code = f.read()
        # Pass constructor variables directly to the contract
        contract = get_contract(contract_code, INITIAL_VALUE)
    return contract


def test_initial_state(adv_storage_contract):
    # Check if the constructor of the contract is set up properly
    assert adv_storage_contract.storedData() == INITIAL_VALUE


def test_failed_transactions(w3, adv_storage_contract, tx_failed):
    k1 = w3.eth.accounts[1]

    # Try to set the storage to a negative amount
    with tx_failed():
        adv_storage_contract.set(-10, transact={"from": k1})

    # Lock the contract by storing more than 100. Then try to change the value
    adv_storage_contract.set(150, transact={"from": k1})
    with tx_failed():
        adv_storage_contract.set(10, transact={"from": k1})

    # Reset the contract and try to change the value
    adv_storage_contract.reset(transact={"from": k1})
    adv_storage_contract.set(10, transact={"from": k1})
    assert adv_storage_contract.storedData() == 10

    # Assert a different exception (ValidationError for non-matching argument type)
    with tx_failed(ValidationError):
        adv_storage_contract.set("foo", transact={"from": k1})

    # Assert a different exception that contains specific text
    with tx_failed(ValidationError, "invocation failed due to improper number of arguments"):
        adv_storage_contract.set(1, 2, transact={"from": k1})


def test_events(w3, adv_storage_contract, get_logs):
    k1, k2 = w3.eth.accounts[:2]

    tx1 = adv_storage_contract.set(10, transact={"from": k1})
    tx2 = adv_storage_contract.set(20, transact={"from": k2})
    tx3 = adv_storage_contract.reset(transact={"from": k1})

    # Save DataChange logs from all three transactions
    logs1 = get_logs(tx1, adv_storage_contract, "DataChange")
    logs2 = get_logs(tx2, adv_storage_contract, "DataChange")
    logs3 = get_logs(tx3, adv_storage_contract, "DataChange")

    # Check log contents
    assert len(logs1) == 1
    assert logs1[0].args.value == 10

    assert len(logs2) == 1
    assert logs2[0].args.setter == k2

    assert not logs3  # tx3 does not generate a log



import pytest


def test_simple_export(make_input_bundle, get_contract):
    lib1 = """
@external
def foo() -> uint256:
    return 5
    """
    main = """
import lib1

exports: lib1.foo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.foo() == 5


def test_export_with_state(make_input_bundle, get_contract):
    lib1 = """
counter: uint256

@external
def foo() -> uint256:
    return self.counter
    """
    main = """
import lib1

initializes: lib1
exports: lib1.foo

@deploy
def __init__():
    lib1.counter = 99
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.foo() == 99


def test_variable_decl_exports(make_input_bundle, get_contract):
    lib1 = """
counter: public(uint256)
FOO: public(immutable(uint256))
BAR: public(constant(uint256)) = 3

@deploy
def __init__():
    self.counter = 1
    FOO = 2
    """
    main = """
import lib1

initializes: lib1
exports: (
    lib1.counter,
    lib1.FOO,
    lib1.BAR,
)

@deploy
def __init__():
    lib1.__init__()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.counter() == 1
    assert c.FOO() == 2
    assert c.BAR() == 3


def test_not_exported(make_input_bundle, get_contract):
    # test that non-exported functions are not in the selector table
    lib1 = """
@external
def foo() -> uint256:
    return 100

@external
def bar() -> uint256:
    return 101
    """
    main = """
import lib1

exports: lib1.foo

@external
def __default__() -> uint256:
    return 127
    """
    caller_code = """
interface Foo:
    def foo() -> uint256: nonpayable
    def bar() -> uint256: nonpayable

@external
def call_bar(foo: Foo) -> uint256:
    return extcall foo.bar()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    c = get_contract(main, input_bundle=input_bundle)
    caller = get_contract(caller_code)

    assert caller.call_bar(c.address) == 127  # default return value


def test_nested_export(make_input_bundle, get_contract):
    lib1 = """
@external
def foo() -> uint256:
    return 5
    """
    lib2 = """
import lib1
    """
    main = """
import lib2

exports: lib2.lib1.foo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.foo() == 5


# not sure if this one should work
@pytest.mark.xfail(reason="ambiguous spec")
def test_recursive_export(make_input_bundle, get_contract):
    lib1 = """
@external
def foo() -> uint256:
    return 5
    """
    lib2 = """
import lib1
exports: lib1.foo
    """
    main = """
import lib2

exports: lib2.foo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.foo() == 5



def test_unreachable_refund(w3, get_contract):
    code = """
@external
def foo():
    assert msg.sender != msg.sender, UNREACHABLE
    """

    c = get_contract(code)
    a0 = w3.eth.accounts[0]
    gas_sent = 10**6
    tx_hash = c.foo(transact={"from": a0, "gas": gas_sent, "gasPrice": 10})
    tx_receipt = w3.eth.get_transaction_receipt(tx_hash)

    assert tx_receipt["status"] == 0
    assert tx_receipt["gasUsed"] == gas_sent  # Drains all gains sent


def test_basic_unreachable(w3, get_contract, tx_failed):
    code = """
@external
def foo(val: int128) -> bool:
    assert val > 0, UNREACHABLE
    assert val == 2, UNREACHABLE
    return True
    """

    c = get_contract(code)

    assert c.foo(2) is True

    with tx_failed(exc_text="Invalid opcode 0xfe"):
        c.foo(1)
    with tx_failed(exc_text="Invalid opcode 0xfe"):
        c.foo(-1)
    with tx_failed(exc_text="Invalid opcode 0xfe"):
        c.foo(-2)


def test_basic_call_unreachable(w3, get_contract, tx_failed):
    code = """

@view
@internal
def _test_me(val: int128) -> bool:
    return val == 33

@external
def foo(val: int128) -> int128:
    assert self._test_me(val), UNREACHABLE
    return -123
    """

    c = get_contract(code)

    assert c.foo(33) == -123

    with tx_failed(exc_text="Invalid opcode 0xfe"):
        c.foo(1)
    with tx_failed(exc_text="Invalid opcode 0xfe"):
        c.foo(-1)


def test_raise_unreachable(w3, get_contract, tx_failed):
    code = """
@external
def foo():
    raise UNREACHABLE
    """

    c = get_contract(code)

    with tx_failed(exc_text="Invalid opcode 0xfe"):
        c.foo()



import pytest

EXPIRY = 16


@pytest.fixture
def auction_start(w3):
    return w3.eth.get_block("latest").timestamp + 1


@pytest.fixture
def auction_contract(w3, get_contract, auction_start):
    with open("examples/auctions/simple_open_auction.vy") as f:
        contract_code = f.read()
        contract = get_contract(contract_code, *[w3.eth.accounts[0], auction_start, EXPIRY])
    return contract


def test_initial_state(w3, tester, auction_contract, auction_start):
    # Check beneficiary is correct
    assert auction_contract.beneficiary() == w3.eth.accounts[0]
    # Check start time is `auction_start`
    assert auction_contract.auctionStart() == auction_start
    # Check time difference between start time and end time is EXPIRY
    assert auction_contract.auctionEnd() == auction_contract.auctionStart() + EXPIRY
    # Check auction has not ended
    assert auction_contract.ended() is False
    # Check highest bidder is empty
    assert auction_contract.highestBidder() is None
    # Check highest bid is 0
    assert auction_contract.highestBid() == 0
    # Check end time is more than current block timestamp
    assert auction_contract.auctionEnd() >= tester.get_block_by_number("latest")["timestamp"]


def test_bid(w3, tester, auction_contract, tx_failed):
    k1, k2, k3, k4, k5 = w3.eth.accounts[:5]
    # Bidder cannot bid 0
    with tx_failed():
        auction_contract.bid(transact={"value": 0, "from": k1})
    # Bidder can bid
    auction_contract.bid(transact={"value": 1, "from": k1})
    # Check that highest bidder and highest bid have changed accordingly
    assert auction_contract.highestBidder() == k1
    assert auction_contract.highestBid() == 1
    # Bidder bid cannot equal current highest bid
    with tx_failed():
        auction_contract.bid(transact={"value": 1, "from": k1})
    # Higher bid can replace current highest bid
    auction_contract.bid(transact={"value": 2, "from": k2})
    # Check that highest bidder and highest bid have changed accordingly
    assert auction_contract.highestBidder() == k2
    assert auction_contract.highestBid() == 2
    # Multiple bidders can bid
    auction_contract.bid(transact={"value": 3, "from": k3})
    auction_contract.bid(transact={"value": 4, "from": k4})
    auction_contract.bid(transact={"value": 5, "from": k5})
    # Check that highest bidder and highest bid have changed accordingly
    assert auction_contract.highestBidder() == k5
    assert auction_contract.highestBid() == 5
    auction_contract.bid(transact={"value": 1 * 10**10, "from": k1})
    pending_return_before_outbid = auction_contract.pendingReturns(k1)
    auction_contract.bid(transact={"value": 2 * 10**10, "from": k2})
    pending_return_after_outbid = auction_contract.pendingReturns(k1)
    # Account has a greater pending return balance after being outbid
    assert pending_return_after_outbid > pending_return_before_outbid

    balance_before_withdrawal = w3.eth.get_balance(k1)
    auction_contract.withdraw(transact={"from": k1})
    balance_after_withdrawal = w3.eth.get_balance(k1)
    # Balance increases after withdrawal
    assert balance_after_withdrawal > balance_before_withdrawal
    # Pending return balance is reset to 0
    assert auction_contract.pendingReturns(k1) == 0


def test_end_auction(w3, tester, auction_contract, tx_failed):
    k1, k2, k3, k4, k5 = w3.eth.accounts[:5]
    # Fails if auction end time has not been reached
    with tx_failed():
        auction_contract.endAuction()
    auction_contract.bid(transact={"value": 1 * 10**10, "from": k2})
    # Move block timestamp forward to reach auction end time
    # tester.time_travel(tester.get_block_by_number('latest')['timestamp'] + EXPIRY)
    w3.testing.mine(EXPIRY)
    balance_before_end = w3.eth.get_balance(k1)
    auction_contract.endAuction(transact={"from": k2})
    balance_after_end = w3.eth.get_balance(k1)
    # Beneficiary receives the highest bid
    assert balance_after_end == balance_before_end + 1 * 10**10
    # Bidder cannot bid after auction end time has been reached
    with tx_failed():
        auction_contract.bid(transact={"value": 10, "from": k1})
    # Auction cannot be ended twice
    with tx_failed():
        auction_contract.endAuction()



import pytest


@pytest.fixture
def c(w3, get_contract):
    with open("examples/crowdfund.vy") as f:
        contract_code = f.read()
        contract = get_contract(contract_code, *[w3.eth.accounts[1], 50, 60])
    return contract


def test_crowdfund_example(c, w3):
    a0, a1, a2, a3, a4, a5, a6 = w3.eth.accounts[:7]
    c.participate(transact={"value": 5})

    assert c.timelimit() == 60
    assert c.deadline() - w3.eth.get_block("latest").timestamp == 59
    assert not w3.eth.get_block("latest").timestamp >= c.deadline()  # expired
    assert not w3.eth.get_balance(c.address) >= c.goal()  # not reached
    c.participate(transact={"value": 49})
    # assert c.reached()
    pre_bal = w3.eth.get_balance(a1)
    w3.testing.mine(100)
    assert not w3.eth.get_block("latest").number >= c.deadline()  # expired
    c.finalize(transact={})
    post_bal = w3.eth.get_balance(a1)
    assert post_bal - pre_bal == 54


def test_crowdfund_example2(c, w3, tx_failed):
    a0, a1, a2, a3, a4, a5, a6 = w3.eth.accounts[:7]
    c.participate(transact={"value": 1, "from": a3})
    c.participate(transact={"value": 2, "from": a4})
    c.participate(transact={"value": 3, "from": a5})
    c.participate(transact={"value": 4, "from": a6})

    assert c.timelimit() == 60
    w3.testing.mine(100)
    # assert c.expired()
    # assert not c.reached()
    pre_bals = [w3.eth.get_balance(x) for x in [a3, a4, a5, a6]]
    with tx_failed():
        c.refund(transact={"from": a0})
    c.refund(transact={"from": a3})
    with tx_failed():
        c.refund(transact={"from": a3})
    c.refund(transact={"from": a4})
    c.refund(transact={"from": a5})
    c.refund(transact={"from": a6})
    post_bals = [w3.eth.get_balance(x) for x in [a3, a4, a5, a6]]
    assert [y - x for x, y in zip(pre_bals, post_bals)] == [1, 2, 3, 4]



from vyper.exceptions import CompilerPanic
from vyper.utils import OrderedSet
from vyper.venom.basicblock import (
    BB_TERMINATORS,
    CFG_ALTERING_INSTRUCTIONS,
    IRBasicBlock,
    IRInstruction,
    IRVariable,
)
from vyper.venom.function import IRFunction


def calculate_cfg(ctx: IRFunction) -> None:
    """
    Calculate (cfg) inputs for each basic block.
    """
    for bb in ctx.basic_blocks:
        bb.cfg_in = OrderedSet()
        bb.cfg_out = OrderedSet()
        bb.out_vars = OrderedSet()

    for bb in ctx.basic_blocks:
        assert len(bb.instructions) > 0, "Basic block should not be empty"
        last_inst = bb.instructions[-1]
        assert last_inst.opcode in BB_TERMINATORS, f"Last instruction should be a terminator {bb}"

        for inst in bb.instructions:
            if inst.opcode in CFG_ALTERING_INSTRUCTIONS:
                ops = inst.get_label_operands()
                for op in ops:
                    ctx.get_basic_block(op.value).add_cfg_in(bb)

    # Fill in the "out" set for each basic block
    for bb in ctx.basic_blocks:
        for in_bb in bb.cfg_in:
            in_bb.add_cfg_out(bb)


def _reset_liveness(ctx: IRFunction) -> None:
    for bb in ctx.basic_blocks:
        for inst in bb.instructions:
            inst.liveness = OrderedSet()


def _calculate_liveness(bb: IRBasicBlock) -> bool:
    """
    Compute liveness of each instruction in the basic block.
    Returns True if liveness changed
    """
    orig_liveness = bb.instructions[0].liveness.copy()
    liveness = bb.out_vars.copy()
    for instruction in reversed(bb.instructions):
        ops = instruction.get_inputs()

        for op in ops:
            if op in liveness:
                instruction.dup_requirements.add(op)

        liveness = liveness.union(OrderedSet.fromkeys(ops))
        out = instruction.get_outputs()[0] if len(instruction.get_outputs()) > 0 else None
        if out in liveness:
            liveness.remove(out)
        instruction.liveness = liveness

    return orig_liveness != bb.instructions[0].liveness


def _calculate_out_vars(bb: IRBasicBlock) -> bool:
    """
    Compute out_vars of basic block.
    Returns True if out_vars changed
    """
    out_vars = bb.out_vars.copy()
    for out_bb in bb.cfg_out:
        target_vars = input_vars_from(bb, out_bb)
        bb.out_vars = bb.out_vars.union(target_vars)
    return out_vars != bb.out_vars


def calculate_liveness(ctx: IRFunction) -> None:
    _reset_liveness(ctx)
    while True:
        changed = False
        for bb in ctx.basic_blocks:
            changed |= _calculate_out_vars(bb)
            changed |= _calculate_liveness(bb)

        if not changed:
            break


# calculate the input variables into self from source
def input_vars_from(source: IRBasicBlock, target: IRBasicBlock) -> OrderedSet[IRVariable]:
    liveness = target.instructions[0].liveness.copy()
    assert isinstance(liveness, OrderedSet)

    for inst in target.instructions:
        if inst.opcode == "phi":
            # we arbitrarily choose one of the arguments to be in the
            # live variables set (dependent on how we traversed into this
            # basic block). the argument will be replaced by the destination
            # operand during instruction selection.
            # for instance, `%56 = phi %label1 %12 %label2 %14`
            # will arbitrarily choose either %12 or %14 to be in the liveness
            # set, and then during instruction selection, after this instruction,
            # %12 will be replaced by %56 in the liveness set
            source1, source2 = inst.operands[0], inst.operands[2]
            phi1, phi2 = inst.operands[1], inst.operands[3]
            if source.label == source1:
                liveness.add(phi1)
                if phi2 in liveness:
                    liveness.remove(phi2)
            elif source.label == source2:
                liveness.add(phi2)
                if phi1 in liveness:
                    liveness.remove(phi1)
            else:
                # bad path into this phi node
                raise CompilerPanic(f"unreachable: {inst}")

    return liveness


# DataFlow Graph
# this could be refactored into its own file, but it's only used here
# for now
class DFG:
    _dfg_inputs: dict[IRVariable, list[IRInstruction]]
    _dfg_outputs: dict[IRVariable, IRInstruction]

    def __init__(self):
        self._dfg_inputs = dict()
        self._dfg_outputs = dict()

    # return uses of a given variable
    def get_uses(self, op: IRVariable) -> list[IRInstruction]:
        return self._dfg_inputs.get(op, [])

    # the instruction which produces this variable.
    def get_producing_instruction(self, op: IRVariable) -> IRInstruction:
        return self._dfg_outputs[op]

    @classmethod
    def build_dfg(cls, ctx: IRFunction) -> "DFG":
        dfg = cls()

        # Build DFG

        # %15 = add %13 %14
        # %16 = iszero %15
        # dfg_outputs of %15 is (%15 = add %13 %14)
        # dfg_inputs of %15 is all the instructions which *use* %15, ex. [(%16 = iszero %15), ...]
        for bb in ctx.basic_blocks:
            for inst in bb.instructions:
                operands = inst.get_inputs()
                res = inst.get_outputs()

                for op in operands:
                    inputs = dfg._dfg_inputs.setdefault(op, [])
                    inputs.append(inst)

                for op in res:  # type: ignore
                    dfg._dfg_outputs[op] = inst

        return dfg



def test_import_interface_types(make_input_bundle, get_contract):
    ifaces = """
interface IFoo:
    def foo() -> uint256: nonpayable
    """

    foo_impl = """
import ifaces

implements: ifaces.IFoo

@external
def foo() -> uint256:
    return block.number
    """

    contract = """
import ifaces

@external
def test_foo(s: ifaces.IFoo) -> bool:
    assert extcall s.foo() == block.number
    return True
    """

    input_bundle = make_input_bundle({"ifaces.vy": ifaces})

    foo = get_contract(foo_impl, input_bundle=input_bundle)
    c = get_contract(contract, input_bundle=input_bundle)

    assert c.test_foo(foo.address) is True



from vyper import ast as vy_ast


def test_compare_different_node_clases():
    vyper_ast = vy_ast.parse_to_ast("foo = 42")
    left = vyper_ast.body[0].target
    right = vyper_ast.body[0].value

    assert left != right
    assert not vy_ast.compare_nodes(left, right)


def test_compare_different_nodes_same_class():
    vyper_ast = vy_ast.parse_to_ast("[1, 2]")
    left, right = vyper_ast.body[0].value.elements

    assert left != right
    assert not vy_ast.compare_nodes(left, right)


def test_compare_different_nodes_same_value():
    vyper_ast = vy_ast.parse_to_ast("[1, 1]")
    left, right = vyper_ast.body[0].value.elements

    assert left != right
    assert vy_ast.compare_nodes(left, right)


def test_compare_complex_nodes_same_value():
    vyper_ast = vy_ast.parse_to_ast("[{'foo':'bar', 43:[1,2,3]}, {'foo':'bar', 43:[1,2,3]}]")
    left, right = vyper_ast.body[0].value.elements

    assert left != right
    assert vy_ast.compare_nodes(left, right)


def test_compare_same_node():
    vyper_ast = vy_ast.parse_to_ast("42")
    node = vyper_ast.body[0].value

    assert node == node
    assert vy_ast.compare_nodes(node, node)



import pytest

from vyper.codegen.context import Context
from vyper.codegen.core import get_type_for_exact_size


class ContextMock(Context):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mock_vars = False
        self._size = 0

    def internal_memory_scope(self):
        if not self._mock_vars:
            for i in range(20):
                self._new_variable(
                    f"#mock{i}", get_type_for_exact_size(self._size), self._size, bool(i % 2)
                )
            self._mock_vars = True
        return super().internal_memory_scope()

    @classmethod
    def set_mock_var_size(cls, size):
        cls._size = size * 32


def pytest_addoption(parser):
    parser.addoption("--memorymock", action="store_true", help="Run tests with mock allocated vars")


def pytest_generate_tests(metafunc):
    if "memory_mocker" in metafunc.fixturenames:
        params = range(1, 11, 2) if metafunc.config.getoption("memorymock") else [False]
        metafunc.parametrize("memory_mocker", params, indirect=True)


def pytest_collection_modifyitems(items, config):
    if config.getoption("memorymock"):
        for item in list(items):
            if "memory_mocker" not in item.fixturenames:
                items.remove(item)

        # hacky magic to ensure the correct number of tests is shown in collection report
        config.pluginmanager.get_plugin("terminalreporter")._numcollected = len(items)


@pytest.fixture
def memory_mocker(monkeypatch, request):
    if request.param:
        monkeypatch.setattr("vyper.codegen.context.Context", ContextMock)
        ContextMock.set_mock_var_size(request.param)



from vyper.compiler import compile_code
from vyper.compiler.output import _compress_source_map
from vyper.compiler.utils import expand_source_map

TEST_CODE = """
@internal
def _baz(a: int128) -> int128:
    b: int128 = a
    for i: int128 in range(2, 5):
        b *=  i
        if b > 31337:
            break
    return b

@internal
def _bar(a: uint256) -> bool:
    if a > 42:
        return True
    return False

@external
def foo(a: uint256) -> int128:
    if self._bar(a):
        return self._baz(2)
    else:
        return 42
    """


def test_jump_map():
    source_map = compile_code(TEST_CODE, output_formats=["source_map"])["source_map"]
    pos_map = source_map["pc_pos_map"]
    jump_map = source_map["pc_jump_map"]

    assert len([v for v in jump_map.values() if v == "o"]) == 1
    assert len([v for v in jump_map.values() if v == "i"]) == 2

    code_lines = [i + "\n" for i in TEST_CODE.split("\n")]
    for pc in [k for k, v in jump_map.items() if v == "o"]:
        lineno, col_offset, _, end_col_offset = pos_map[pc]
        assert code_lines[lineno - 1][col_offset:end_col_offset].startswith("return")

    for pc in [k for k, v in jump_map.items() if v == "i"]:
        lineno, col_offset, _, end_col_offset = pos_map[pc]
        assert code_lines[lineno - 1][col_offset:end_col_offset].startswith("self.")


def test_pos_map_offsets():
    source_map = compile_code(TEST_CODE, output_formats=["source_map"])["source_map"]
    expanded = expand_source_map(source_map["pc_pos_map_compressed"])

    pc_iter = iter(source_map["pc_pos_map"][i] for i in sorted(source_map["pc_pos_map"]))
    jump_iter = iter(source_map["pc_jump_map"][i] for i in sorted(source_map["pc_jump_map"]))
    code_lines = [i + "\n" for i in TEST_CODE.split("\n")]

    for item in expanded:
        if item[-1] is not None:
            assert next(jump_iter) == item[-1]

        if item[:2] != [-1, -1]:
            start, length = item[:2]
            lineno, col_offset, end_lineno, end_col_offset = next(pc_iter)
            assert code_lines[lineno - 1][col_offset] == TEST_CODE[start]
            assert length == (
                sum(len(i) for i in code_lines[lineno - 1 : end_lineno])
                - col_offset
                - (len(code_lines[end_lineno - 1]) - end_col_offset)
            )


def test_error_map():
    code = """
foo: uint256

@external
def update_foo():
    self.foo += 1
    """
    error_map = compile_code(code, output_formats=["source_map"])["source_map"]["error_map"]
    assert "safeadd" in list(error_map.values())
    assert "fallback function" in list(error_map.values())


def test_compress_source_map():
    code = """
@external
def foo() -> uint256:
    return 42
    """
    compressed = _compress_source_map(
        code, {"0": None, "2": (2, 0, 4, 13), "3": (2, 0, 2, 8), "5": (2, 0, 2, 8)}, {"3": "o"}, 2
    )
    assert compressed == "-1:-1:2:-;1:45;:8::o;"


def test_expand_source_map():
    compressed = "-1:-1:0:-;;13:42:1;:21;::0:o;:::-;1::1;"
    expanded = [
        [-1, -1, 0, "-"],
        [-1, -1, 0, None],
        [13, 42, 1, None],
        [13, 21, 1, None],
        [13, 21, 0, "o"],
        [13, 21, 0, "-"],
        [1, 21, 1, None],
    ]
    assert expand_source_map(compressed) == expanded



import pytest
from eth_tester.exceptions import TransactionFailed


# web3 returns f"execution reverted: {err_str}"
# TODO move exception string parsing logic into tx_failed
def _fixup_err_str(s):
    return s.replace("execution reverted: ", "")


def test_assert_refund(w3, get_contract_with_gas_estimation, tx_failed):
    code = """
@external
def foo():
    raise
    """
    c = get_contract_with_gas_estimation(code)
    a0 = w3.eth.accounts[0]
    gas_sent = 10**6
    tx_hash = c.foo(transact={"from": a0, "gas": gas_sent, "gasPrice": 10})
    # More info on receipt status:
    # https://github.com/ethereum/EIPs/blob/master/EIPS/eip-658.md#specification.
    tx_receipt = w3.eth.get_transaction_receipt(tx_hash)
    assert tx_receipt["status"] == 0
    # Checks for gas refund from revert
    assert tx_receipt["gasUsed"] < gas_sent


def test_assert_reason(w3, get_contract_with_gas_estimation, tx_failed, memory_mocker):
    code = """
@external
def test(a: int128) -> int128:
    assert a > 1, "larger than one please"
    return 1 + a

@external
def test2(a: int128, b: int128, extra_reason: String[32]) -> int128:
    c: int128 = 11
    assert a > 1, "a is not large enough"
    assert b == 1, concat("b may only be 1", extra_reason)
    return a + b + c

@external
def test3(reason_str: String[32]):
    raise reason_str
    """
    c = get_contract_with_gas_estimation(code)

    assert c.test(2) == 3
    with pytest.raises(TransactionFailed) as e_info:
        c.test(0)

    assert _fixup_err_str(e_info.value.args[0]) == "larger than one please"
    # a = 0, b = 1
    with pytest.raises(TransactionFailed) as e_info:
        c.test2(0, 1, "")
    assert _fixup_err_str(e_info.value.args[0]) == "a is not large enough"
    # a = 1, b = 0
    with pytest.raises(TransactionFailed) as e_info:
        c.test2(2, 2, " because I said so")
    assert _fixup_err_str(e_info.value.args[0]) == "b may only be 1" + " because I said so"
    # return correct value
    assert c.test2(5, 1, "") == 17

    with pytest.raises(TransactionFailed) as e_info:
        c.test3("An exception")
    assert _fixup_err_str(e_info.value.args[0]) == "An exception"


invalid_code = [
    """
@external
def test(a: int128) -> int128:
    assert a > 1, ""
    return 1 + a
    """,
    """
@external
def test(a: int128) -> int128:
    raise ""
    """,
    """
@external
def test():
    assert create_minimal_proxy_to(self)
    """,
]


@pytest.mark.parametrize("code", invalid_code)
def test_invalid_assertions(get_contract, assert_compile_failed, code):
    assert_compile_failed(lambda: get_contract(code))


valid_code = [
    """
@external
def mint(_to: address, _value: uint256):
    raise
    """,
    """
@internal
def ret1() -> int128:
    return 1
@external
def test():
    assert self.ret1() == 1
    """,
    """
@external
def test():
    assert raw_call(msg.sender, b'', max_outsize=1, gas=10, value=1000*1000) == b''
    """,
    """
@external
def test():
    assert create_minimal_proxy_to(self) == self
    """,
]


@pytest.mark.parametrize("code", valid_code)
def test_valid_assertions(get_contract, code):
    get_contract(code)


def test_assert_staticcall(get_contract, tx_failed, memory_mocker):
    foreign_code = """
state: uint256
@external
def not_really_constant() -> uint256:
    self.state += 1
    return self.state
    """
    code = """
interface ForeignContract:
    def not_really_constant() -> uint256: view

@external
def test():
    assert staticcall ForeignContract(msg.sender).not_really_constant() == 1
    """
    c1 = get_contract(foreign_code)
    c2 = get_contract(code, *[c1.address])
    # static call prohibits state change
    with tx_failed():
        c2.test()


def test_assert_in_for_loop(get_contract, tx_failed, memory_mocker):
    code = """
@external
def test(x: uint256[3]) -> bool:
    for i: uint256 in range(3):
        assert x[i] < 5
    return True
    """

    c = get_contract(code)

    c.test([1, 2, 3])
    with tx_failed():
        c.test([5, 1, 3])
    with tx_failed():
        c.test([1, 5, 3])
    with tx_failed():
        c.test([1, 3, 5])


def test_assert_with_reason_in_for_loop(get_contract, tx_failed, memory_mocker):
    code = """
@external
def test(x: uint256[3]) -> bool:
    for i: uint256 in range(3):
        assert x[i] < 5, "because reasons"
    return True
    """

    c = get_contract(code)

    c.test([1, 2, 3])
    with tx_failed():
        c.test([5, 1, 3])
    with tx_failed():
        c.test([1, 5, 3])
    with tx_failed():
        c.test([1, 3, 5])


def test_assert_reason_revert_length(w3, get_contract, tx_failed, memory_mocker):
    code = """
@external
def test() -> int128:
    assert 1 == 2, "oops"
    return 1
"""
    c = get_contract(code)
    with tx_failed(exc_text="oops"):
        c.test()



import pytest

MAX_BIDS = 128

BIDDING_TIME = 150
REVEAL_TIME = 50
TEST_INCREMENT = 1


@pytest.fixture
def auction_contract(w3, get_contract):
    with open("examples/auctions/blind_auction.vy") as f:
        contract_code = f.read()
        contract = get_contract(contract_code, *[w3.eth.accounts[0], BIDDING_TIME, REVEAL_TIME])
    return contract


def test_initial_state(w3, tester, auction_contract):
    # Check beneficiary is correct
    assert auction_contract.beneficiary() == w3.eth.accounts[0]
    # Check that bidding end time is correct
    assert (
        auction_contract.biddingEnd()
        == tester.get_block_by_number("latest")["timestamp"] + BIDDING_TIME
    )  # noqa: E501
    # Check that the reveal end time is correct
    assert auction_contract.revealEnd() == auction_contract.biddingEnd() + REVEAL_TIME
    # Check auction has not ended
    assert auction_contract.ended() is False
    # Check highest bid is 0
    assert auction_contract.highestBid() == 0
    # Check highest bidder is empty
    assert auction_contract.highestBidder() is None


def test_late_bid(w3, auction_contract, tx_failed):
    k1 = w3.eth.accounts[1]

    # Move time forward past bidding end
    w3.testing.mine(BIDDING_TIME + TEST_INCREMENT)

    # Try to bid after bidding has ended
    with tx_failed():
        auction_contract.bid(
            w3.keccak(
                b"".join(
                    [
                        (200).to_bytes(32, byteorder="big"),
                        (0).to_bytes(32, byteorder="big"),
                        (8675309).to_bytes(32, byteorder="big"),
                    ]
                )
            ),
            transact={"value": 200, "from": k1},
        )


def test_too_many_bids(w3, auction_contract, tx_failed):
    k1 = w3.eth.accounts[1]

    # First 128 bids should be able to be placed successfully
    for i in range(MAX_BIDS):
        auction_contract.bid(
            w3.keccak(
                b"".join(
                    [
                        (i).to_bytes(32, byteorder="big"),
                        (1).to_bytes(32, byteorder="big"),
                        (8675309).to_bytes(32, byteorder="big"),
                    ]
                )
            ),
            transact={"value": i, "from": k1},
        )

    # 129th bid should fail
    with tx_failed():
        auction_contract.bid(
            w3.keccak(
                b"".join(
                    [
                        (128).to_bytes(32, byteorder="big"),
                        (0).to_bytes(32, byteorder="big"),
                        (8675309).to_bytes(32, byteorder="big"),
                    ]
                )
            ),
            transact={"value": 128, "from": k1},
        )


def test_early_reval(w3, auction_contract, tx_failed):
    k1 = w3.eth.accounts[1]

    # k1 places 1 real bid
    auction_contract.bid(
        w3.keccak(
            b"".join(
                [
                    (100).to_bytes(32, byteorder="big"),
                    (0).to_bytes(32, byteorder="big"),
                    (8675309).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        transact={"value": 100, "from": k1},
    )

    # Move time slightly forward (still before bidding has ended)
    w3.testing.mine(TEST_INCREMENT)

    # Try to reveal early
    _values = [0] * MAX_BIDS  # Initialized with 128 default values
    _fakes = [False] * MAX_BIDS  # Initialized with 128 default values
    _secrets = [b"\x00" * 32] * MAX_BIDS  # Initialized with 128 default values
    _numBids = 1
    _values[0] = 100
    _fakes[0] = False
    _secrets[0] = (8675309).to_bytes(32, byteorder="big")
    with tx_failed():
        auction_contract.reveal(
            _numBids, _values, _fakes, _secrets, transact={"value": 0, "from": k1}
        )

    # Check highest bidder is still empty
    assert auction_contract.highestBidder() is None
    # Check highest bid is still 0
    assert auction_contract.highestBid() == 0


def test_late_reveal(w3, auction_contract, tx_failed):
    k1 = w3.eth.accounts[1]

    # k1 places 1 real bid
    auction_contract.bid(
        w3.keccak(
            b"".join(
                [
                    (100).to_bytes(32, byteorder="big"),
                    (0).to_bytes(32, byteorder="big"),
                    (8675309).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        transact={"value": 100, "from": k1},
    )

    # Move time forward past bidding _and_ reveal time
    w3.testing.mine(BIDDING_TIME + REVEAL_TIME + TEST_INCREMENT)

    # Try to reveal late
    _values = [0] * MAX_BIDS  # Initialized with 128 default values
    _fakes = [False] * MAX_BIDS  # Initialized with 128 default values
    _secrets = [b"\x00" * 32] * MAX_BIDS  # Initialized with 128 default values
    _numBids = 1
    _values[0] = 100
    _fakes[0] = False
    _secrets[0] = (8675309).to_bytes(32, byteorder="big")
    with tx_failed():
        auction_contract.reveal(
            _numBids, _values, _fakes, _secrets, transact={"value": 0, "from": k1}
        )

    # Check highest bidder is still empty
    assert auction_contract.highestBidder() is None
    # Check highest bid is still 0
    assert auction_contract.highestBid() == 0


def test_early_end(w3, auction_contract, tx_failed):
    k0 = w3.eth.accounts[0]

    # Should not be able to end auction before reveal time has ended
    with tx_failed():
        auction_contract.auctionEnd(transact={"value": 0, "from": k0})


def test_double_end(w3, auction_contract, tx_failed):
    k0 = w3.eth.accounts[0]

    # Move time forward past bidding and reveal end
    w3.testing.mine(BIDDING_TIME + REVEAL_TIME + TEST_INCREMENT)

    # First auction end should succeed
    auction_contract.auctionEnd(transact={"value": 0, "from": k0})

    # Should not be able to end auction twice
    with tx_failed():
        auction_contract.auctionEnd(transact={"value": 0, "from": k0})


def test_blind_auction(w3, auction_contract):
    k0, k1, k2, k3 = w3.eth.accounts[0:4]

    ###################################################################
    #                         Place bids                              #
    ###################################################################

    # k1 places 1 real bid
    auction_contract.bid(
        w3.keccak(
            b"".join(
                [
                    (100).to_bytes(32, byteorder="big"),
                    (0).to_bytes(32, byteorder="big"),
                    (8675309).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        transact={"value": 100, "from": k1},
    )

    # k2 places 1 real bid (highest) and 2 fake
    auction_contract.bid(
        w3.keccak(
            b"".join(
                [
                    (150).to_bytes(32, byteorder="big"),
                    (1).to_bytes(32, byteorder="big"),
                    (1234567).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        transact={"value": 150, "from": k2},
    )
    auction_contract.bid(
        w3.keccak(
            b"".join(
                [
                    (200).to_bytes(32, byteorder="big"),
                    (0).to_bytes(32, byteorder="big"),
                    (1234567).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        transact={"value": 250, "from": k2},
    )
    auction_contract.bid(
        w3.keccak(
            b"".join(
                [
                    (300).to_bytes(32, byteorder="big"),
                    (1).to_bytes(32, byteorder="big"),
                    (1234567).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        transact={"value": 300, "from": k2},
    )

    # k3 places 2 fake bids
    auction_contract.bid(
        w3.keccak(
            b"".join(
                [
                    (175).to_bytes(32, byteorder="big"),
                    (1).to_bytes(32, byteorder="big"),
                    (9876543).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        transact={"value": 175, "from": k3},
    )
    auction_contract.bid(
        w3.keccak(
            b"".join(
                [
                    (275).to_bytes(32, byteorder="big"),
                    (1).to_bytes(32, byteorder="big"),
                    (9876543).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        transact={"value": 275, "from": k3},
    )

    ###################################################################
    #                          Reveal bids.                           #
    ###################################################################

    # Move time forward past bidding end (still within reveal end)
    w3.testing.mine(BIDDING_TIME + TEST_INCREMENT)

    # Reveal k1 bids
    _values = [0] * MAX_BIDS  # Initialized with 128 default values
    _fakes = [False] * MAX_BIDS  # Initialized with 128 default values
    _secrets = [b"\x00" * 32] * MAX_BIDS  # Initialized with 128 default values
    _numBids = 1
    _values[0] = 100
    _fakes[0] = False
    _secrets[0] = (8675309).to_bytes(32, byteorder="big")
    auction_contract.reveal(_numBids, _values, _fakes, _secrets, transact={"value": 0, "from": k1})

    #: Check that highest bidder and highest bid have updated
    assert auction_contract.highestBid() == 100
    assert auction_contract.highestBidder() == k1

    # Reveal k2 bids
    _values = [0] * MAX_BIDS  # Initialized with 128 default values
    _fakes = [False] * MAX_BIDS  # Initialized with 128 default values
    _secrets = [b"\x00" * 32] * MAX_BIDS  # Initialized with 128 default values
    _values[0] = 150
    _fakes[0] = True
    _secrets[0] = (1234567).to_bytes(32, byteorder="big")
    _values[1] = 200
    _fakes[1] = False
    _secrets[1] = (1234567).to_bytes(32, byteorder="big")
    _values[2] = 300
    _fakes[2] = True
    _secrets[2] = (1234567).to_bytes(32, byteorder="big")
    balance_before_reveal = w3.eth.get_balance(k2)
    auction_contract.reveal(3, _values, _fakes, _secrets, transact={"value": 0, "from": k2})
    balance_after_reveal = w3.eth.get_balance(k2)

    #: Check that highest bidder and highest bid have updated
    assert auction_contract.highestBid() == 200
    assert auction_contract.highestBidder() == k2

    # Check for refund from false bids (300 + 150) and partial deposit (50)
    assert balance_after_reveal == (balance_before_reveal + 500)

    # Reveal k3 bids
    _values = [0] * MAX_BIDS  # Initialized with 128 default values
    _fakes = [False] * MAX_BIDS  # Initialized with 128 default values
    _secrets = [b"\x00" * 32] * MAX_BIDS  # Initialized with 128 default values
    _values[0] = 175
    _fakes[0] = True
    _secrets[0] = (9876543).to_bytes(32, byteorder="big")
    _values[1] = 275
    _fakes[1] = True
    _secrets[1] = (9876543).to_bytes(32, byteorder="big")
    balance_before_reveal = w3.eth.get_balance(k3)
    auction_contract.reveal(2, _values, _fakes, _secrets, transact={"value": 0, "from": k3})
    balance_after_reveal = w3.eth.get_balance(k3)

    #: Check that highest bidder and highest bid have NOT updated
    assert auction_contract.highestBidder() == k2
    assert auction_contract.highestBid() == 200

    # Check for refund from false bids (175 + 275 = 450)
    assert balance_after_reveal == (balance_before_reveal + 450)

    ###################################################################
    #                          End auction.                           #
    ###################################################################

    # Move time forward past bidding and reveal end
    w3.testing.mine(REVEAL_TIME)

    # End the auction
    balance_before_end = w3.eth.get_balance(k0)
    auction_contract.auctionEnd(transact={"value": 0, "from": k0})
    balance_after_end = w3.eth.get_balance(k0)

    # Check that auction indeed ended
    assert auction_contract.ended() is True

    # Check that beneficiary has been paid
    assert balance_after_end == (balance_before_end + 200)

    # Check that k1 is able to withdraw their outbid bid
    balance_before_withdraw = w3.eth.get_balance(k1)
    auction_contract.withdraw(transact={"value": 0, "from": k1})
    balance_after_withdraw = w3.eth.get_balance(k1)
    assert balance_after_withdraw == (balance_before_withdraw + 100)



from vyper.utils import ir_pass
from vyper.venom.basicblock import IRBasicBlock
from vyper.venom.function import IRFunction


def _process_basic_block(ctx: IRFunction, bb: IRBasicBlock):
    pass


@ir_pass
def ir_pass_constant_propagation(ctx: IRFunction):
    for bb in ctx.basic_blocks:
        _process_basic_block(ctx, bb)



def test_gas_call(get_contract_with_gas_estimation):
    gas_call = """
@external
def foo() -> uint256:
    return msg.gas
    """

    c = get_contract_with_gas_estimation(gas_call)

    assert c.foo(call={"gas": 50000}) < 50000
    assert c.foo(call={"gas": 50000}) > 25000



import pytest

from vyper.compiler import compile_code
from vyper.compiler.settings import Settings
from vyper.evm.opcodes import EVM_VERSIONS
from vyper.exceptions import StructureException

post_cancun = {k: v for k, v in EVM_VERSIONS.items() if v >= EVM_VERSIONS["cancun"]}


@pytest.mark.parametrize("evm_version", list(EVM_VERSIONS.keys()))
def test_transient_blocked(evm_version):
    # test transient is blocked on pre-cancun and compiles post-cancun
    code = """
my_map: transient(HashMap[address, uint256])
    """
    settings = Settings(evm_version=evm_version)
    if EVM_VERSIONS[evm_version] >= EVM_VERSIONS["cancun"]:
        assert compile_code(code, settings=settings) is not None
    else:
        with pytest.raises(StructureException):
            compile_code(code, settings=settings)


@pytest.mark.parametrize("evm_version", list(post_cancun.keys()))
def test_transient_compiles(evm_version):
    # test transient keyword at least generates TLOAD/TSTORE opcodes
    settings = Settings(evm_version=evm_version)
    getter_code = """
my_map: public(transient(HashMap[address, uint256]))
    """
    t = compile_code(getter_code, settings=settings, output_formats=["opcodes_runtime"])
    t = t["opcodes_runtime"].split(" ")

    assert "TLOAD" in t
    assert "TSTORE" not in t

    setter_code = """
my_map: transient(HashMap[address, uint256])

@external
def setter(k: address, v: uint256):
    self.my_map[k] = v
    """
    t = compile_code(setter_code, settings=settings, output_formats=["opcodes_runtime"])
    t = t["opcodes_runtime"].split(" ")

    assert "TLOAD" not in t
    assert "TSTORE" in t

    getter_setter_code = """
my_map: public(transient(HashMap[address, uint256]))

@external
def setter(k: address, v: uint256):
    self.my_map[k] = v
    """
    t = compile_code(getter_setter_code, settings=settings, output_formats=["opcodes_runtime"])
    t = t["opcodes_runtime"].split(" ")

    assert "TLOAD" in t
    assert "TSTORE" in t



def test_constant_address_balance(w3, get_contract_with_gas_estimation):
    code = """
a: constant(address) = 0x776Ba14735FF84789320718cf0aa43e91F7A8Ce1

@external
def foo() -> uint256:
    x: uint256 = a.balance
    return x
    """
    address = "0x776Ba14735FF84789320718cf0aa43e91F7A8Ce1"

    c = get_contract_with_gas_estimation(code)

    assert c.foo() == 0

    w3.eth.send_transaction({"to": address, "value": 1337})

    assert c.foo() == 1337



from vyper.exceptions import FunctionDeclarationException


def test_constant_test(get_contract_with_gas_estimation_for_constants):
    constant_test = """
@external
@view
def foo() -> int128:
    return 5
    """

    c = get_contract_with_gas_estimation_for_constants(constant_test)
    assert c.foo() == 5

    print("Passed constant function test")


def test_invalid_constant_and_payable(
    get_contract_with_gas_estimation_for_constants, assert_compile_failed
):
    code = """
@external
@payable
@view
def foo() -> num:
    return 5
"""
    assert_compile_failed(
        lambda: get_contract_with_gas_estimation_for_constants(code), FunctionDeclarationException
    )



import pytest

from vyper import compiler
from vyper.compiler.settings import Settings
from vyper.evm.opcodes import EVM_VERSIONS


@pytest.mark.parametrize("evm_version", list(EVM_VERSIONS))
def test_self_balance(w3, get_contract_with_gas_estimation, evm_version):
    code = """
@external
@view
def get_balance() -> uint256:
    a: uint256 = self.balance
    return a

@external
@payable
def __default__():
    pass
    """
    settings = Settings(evm_version=evm_version)
    opcodes = compiler.compile_code(code, output_formats=["opcodes"], settings=settings)["opcodes"]
    if EVM_VERSIONS[evm_version] >= EVM_VERSIONS["istanbul"]:
        assert "SELFBALANCE" in opcodes
    else:
        assert "SELFBALANCE" not in opcodes

    c = get_contract_with_gas_estimation(code, evm_version=evm_version)
    w3.eth.send_transaction({"to": c.address, "value": 1337})

    assert c.get_balance() == 1337



from typing import Optional

from vyper.codegen.context import VariableRecord
from vyper.codegen.ir_node import IRnode
from vyper.evm.opcodes import get_opcodes
from vyper.exceptions import CompilerPanic
from vyper.ir.compile_ir import is_mem_sym, is_symbol
from vyper.semantics.types.function import ContractFunctionT
from vyper.utils import MemoryPositions, OrderedSet
from vyper.venom.basicblock import (
    IRBasicBlock,
    IRInstruction,
    IRLabel,
    IRLiteral,
    IROperand,
    IRVariable,
    MemType,
)
from vyper.venom.function import IRFunction

_BINARY_IR_INSTRUCTIONS = frozenset(
    [
        "eq",
        "gt",
        "lt",
        "slt",
        "sgt",
        "shr",
        "shl",
        "or",
        "xor",
        "and",
        "add",
        "sub",
        "mul",
        "div",
        "mod",
        "exp",
        "sha3",
        "sha3_64",
        "signextend",
    ]
)

# Instructions that are mapped to their inverse
INVERSE_MAPPED_IR_INSTRUCTIONS = {"ne": "eq", "le": "gt", "sle": "sgt", "ge": "lt", "sge": "slt"}

# Instructions that have a direct EVM opcode equivalent and can
# be passed through to the EVM assembly without special handling
PASS_THROUGH_INSTRUCTIONS = [
    "chainid",
    "basefee",
    "timestamp",
    "blockhash",
    "caller",
    "selfbalance",
    "calldatasize",
    "callvalue",
    "address",
    "origin",
    "codesize",
    "gas",
    "gasprice",
    "gaslimit",
    "returndatasize",
    "coinbase",
    "number",
    "iszero",
    "not",
    "calldataload",
    "extcodesize",
    "extcodehash",
    "balance",
]

SymbolTable = dict[str, Optional[IROperand]]


def _get_symbols_common(a: dict, b: dict) -> dict:
    ret = {}
    # preserves the ordering in `a`
    for k in a.keys():
        if k not in b:
            continue
        if a[k] == b[k]:
            continue
        ret[k] = a[k], b[k]
    return ret


# convert IRnode directly to venom
def ir_node_to_venom(ir: IRnode) -> IRFunction:
    ctx = IRFunction()
    _convert_ir_bb(ctx, ir, {}, OrderedSet(), {})

    # Patch up basic blocks. Connect unterminated blocks to the next with
    # a jump. terminate final basic block with STOP.
    for i, bb in enumerate(ctx.basic_blocks):
        if not bb.is_terminated:
            if i < len(ctx.basic_blocks) - 1:
                bb.append_instruction("jmp", ctx.basic_blocks[i + 1].label)
            else:
                bb.append_instruction("stop")

    return ctx


def _convert_binary_op(
    ctx: IRFunction,
    ir: IRnode,
    symbols: SymbolTable,
    variables: OrderedSet,
    allocated_variables: dict[str, IRVariable],
    swap: bool = False,
) -> Optional[IRVariable]:
    ir_args = ir.args[::-1] if swap else ir.args
    arg_0, arg_1 = _convert_ir_bb_list(ctx, ir_args, symbols, variables, allocated_variables)

    assert isinstance(ir.value, str)  # mypy hint
    return ctx.get_basic_block().append_instruction(ir.value, arg_1, arg_0)


def _append_jmp(ctx: IRFunction, label: IRLabel) -> None:
    ctx.get_basic_block().append_instruction("jmp", label)

    label = ctx.get_next_label()
    bb = IRBasicBlock(label, ctx)
    ctx.append_basic_block(bb)


def _new_block(ctx: IRFunction) -> IRBasicBlock:
    bb = IRBasicBlock(ctx.get_next_label(), ctx)
    bb = ctx.append_basic_block(bb)
    return bb


def _handle_self_call(
    ctx: IRFunction,
    ir: IRnode,
    symbols: SymbolTable,
    variables: OrderedSet,
    allocated_variables: dict[str, IRVariable],
) -> Optional[IRVariable]:
    func_t = ir.passthrough_metadata.get("func_t", None)
    args_ir = ir.passthrough_metadata["args_ir"]
    goto_ir = [ir for ir in ir.args if ir.value == "goto"][0]
    target_label = goto_ir.args[0].value  # goto
    return_buf = goto_ir.args[1]  # return buffer
    ret_args: list[IROperand] = [IRLabel(target_label)]  # type: ignore

    for arg in args_ir:
        if arg.is_literal:
            sym = symbols.get(f"&{arg.value}", None)
            if sym is None:
                ret = _convert_ir_bb(ctx, arg, symbols, variables, allocated_variables)
                ret_args.append(ret)
            else:
                ret_args.append(sym)  # type: ignore
        else:
            ret = _convert_ir_bb(ctx, arg._optimized, symbols, variables, allocated_variables)
            if arg.location and arg.location.load_op == "calldataload":
                bb = ctx.get_basic_block()
                ret = bb.append_instruction(arg.location.load_op, ret)
            ret_args.append(ret)

    if return_buf.is_literal:
        ret_args.append(return_buf.value)  # type: ignore

    bb = ctx.get_basic_block()
    do_ret = func_t.return_type is not None
    if do_ret:
        invoke_ret = bb.append_invoke_instruction(ret_args, returns=True)  # type: ignore
        allocated_variables["return_buffer"] = invoke_ret  # type: ignore
        return invoke_ret
    else:
        bb.append_invoke_instruction(ret_args, returns=False)  # type: ignore
        return None


def _handle_internal_func(
    ctx: IRFunction, ir: IRnode, func_t: ContractFunctionT, symbols: SymbolTable
) -> IRnode:
    bb = IRBasicBlock(IRLabel(ir.args[0].args[0].value, True), ctx)  # type: ignore
    bb = ctx.append_basic_block(bb)

    old_ir_mempos = 0
    old_ir_mempos += 64

    for arg in func_t.arguments:
        symbols[f"&{old_ir_mempos}"] = bb.append_instruction("param")
        bb.instructions[-1].annotation = arg.name
        old_ir_mempos += 32  # arg.typ.memory_bytes_required

    # return buffer
    if func_t.return_type is not None:
        symbols["return_buffer"] = bb.append_instruction("param")
        bb.instructions[-1].annotation = "return_buffer"

    # return address
    symbols["return_pc"] = bb.append_instruction("param")
    bb.instructions[-1].annotation = "return_pc"

    return ir.args[0].args[2]


def _convert_ir_simple_node(
    ctx: IRFunction,
    ir: IRnode,
    symbols: SymbolTable,
    variables: OrderedSet,
    allocated_variables: dict[str, IRVariable],
) -> Optional[IRVariable]:
    args = [_convert_ir_bb(ctx, arg, symbols, variables, allocated_variables) for arg in ir.args]
    return ctx.get_basic_block().append_instruction(ir.value, *args)  # type: ignore


_break_target: Optional[IRBasicBlock] = None
_continue_target: Optional[IRBasicBlock] = None


def _get_variable_from_address(
    variables: OrderedSet[VariableRecord], addr: int
) -> Optional[VariableRecord]:
    assert isinstance(addr, int), "non-int address"
    for var in variables.keys():
        if var.location.name != "memory":
            continue
        if addr >= var.pos and addr < var.pos + var.size:  # type: ignore
            return var
    return None


def _append_return_for_stack_operand(
    ctx: IRFunction, symbols: SymbolTable, ret_ir: IRVariable, last_ir: IRVariable
) -> None:
    bb = ctx.get_basic_block()
    if isinstance(ret_ir, IRLiteral):
        sym = symbols.get(f"&{ret_ir.value}", None)
        new_var = bb.append_instruction("alloca", 32, ret_ir)
        bb.append_instruction("mstore", sym, new_var)  # type: ignore
    else:
        sym = symbols.get(ret_ir.value, None)
        if sym is None:
            # FIXME: needs real allocations
            new_var = bb.append_instruction("alloca", 32, 0)
            bb.append_instruction("mstore", ret_ir, new_var)  # type: ignore
        else:
            new_var = ret_ir
    bb.append_instruction("return", last_ir, new_var)  # type: ignore


def _convert_ir_bb_list(ctx, ir, symbols, variables, allocated_variables):
    ret = []
    for ir_node in ir:
        venom = _convert_ir_bb(ctx, ir_node, symbols, variables, allocated_variables)
        assert venom is not None, ir_node
        ret.append(venom)
    return ret


def _convert_ir_bb(ctx, ir, symbols, variables, allocated_variables):
    assert isinstance(ir, IRnode), ir
    assert isinstance(variables, OrderedSet)
    global _break_target, _continue_target

    frame_info = ir.passthrough_metadata.get("frame_info", None)
    if frame_info is not None:
        local_vars = OrderedSet[VariableRecord](frame_info.frame_vars.values())
        variables |= local_vars

    assert isinstance(variables, OrderedSet)

    if ir.value in _BINARY_IR_INSTRUCTIONS:
        return _convert_binary_op(
            ctx, ir, symbols, variables, allocated_variables, ir.value in ["sha3_64"]
        )

    elif ir.value in INVERSE_MAPPED_IR_INSTRUCTIONS:
        org_value = ir.value
        ir.value = INVERSE_MAPPED_IR_INSTRUCTIONS[ir.value]
        new_var = _convert_binary_op(ctx, ir, symbols, variables, allocated_variables)
        ir.value = org_value
        return ctx.get_basic_block().append_instruction("iszero", new_var)

    elif ir.value in PASS_THROUGH_INSTRUCTIONS:
        return _convert_ir_simple_node(ctx, ir, symbols, variables, allocated_variables)

    elif ir.value in ["pass", "stop", "return"]:
        pass
    elif ir.value == "deploy":
        ctx.ctor_mem_size = ir.args[0].value
        ctx.immutables_len = ir.args[2].value
        return None
    elif ir.value == "seq":
        func_t = ir.passthrough_metadata.get("func_t", None)
        if ir.is_self_call:
            return _handle_self_call(ctx, ir, symbols, variables, allocated_variables)
        elif func_t is not None:
            symbols = {}
            allocated_variables = {}
            variables = OrderedSet(
                {v: True for v in ir.passthrough_metadata["frame_info"].frame_vars.values()}
            )
            if func_t.is_internal:
                ir = _handle_internal_func(ctx, ir, func_t, symbols)
            # fallthrough

        ret = None
        for ir_node in ir.args:  # NOTE: skip the last one
            ret = _convert_ir_bb(ctx, ir_node, symbols, variables, allocated_variables)

        return ret
    elif ir.value in ["staticcall", "call"]:  # external call
        idx = 0
        gas = _convert_ir_bb(ctx, ir.args[idx], symbols, variables, allocated_variables)
        address = _convert_ir_bb(ctx, ir.args[idx + 1], symbols, variables, allocated_variables)

        value = None
        if ir.value == "call":
            value = _convert_ir_bb(ctx, ir.args[idx + 2], symbols, variables, allocated_variables)
        else:
            idx -= 1

        argsOffset, argsSize, retOffset, retSize = _convert_ir_bb_list(
            ctx, ir.args[idx + 3 : idx + 7], symbols, variables, allocated_variables
        )

        if isinstance(argsOffset, IRLiteral):
            offset = int(argsOffset.value)
            addr = offset - 32 + 4 if offset > 0 else 0
            argsOffsetVar = symbols.get(f"&{addr}", None)
            if argsOffsetVar is None:
                argsOffsetVar = argsOffset
            elif isinstance(argsOffsetVar, IRVariable):
                argsOffsetVar.mem_type = MemType.MEMORY
                argsOffsetVar.mem_addr = addr
                argsOffsetVar.offset = 32 - 4 if offset > 0 else 0
            else:  # pragma: nocover
                raise CompilerPanic("unreachable")
        else:
            argsOffsetVar = argsOffset

        retOffsetValue = int(retOffset.value) if retOffset else 0
        retVar = ctx.get_next_variable(MemType.MEMORY, retOffsetValue)
        symbols[f"&{retOffsetValue}"] = retVar

        bb = ctx.get_basic_block()

        if ir.value == "call":
            args = [retSize, retOffset, argsSize, argsOffsetVar, value, address, gas]
            return bb.append_instruction(ir.value, *args)
        else:
            args = [retSize, retOffset, argsSize, argsOffsetVar, address, gas]
            return bb.append_instruction(ir.value, *args)
    elif ir.value == "if":
        cond = ir.args[0]

        # convert the condition
        cont_ret = _convert_ir_bb(ctx, cond, symbols, variables, allocated_variables)
        current_bb = ctx.get_basic_block()

        else_block = IRBasicBlock(ctx.get_next_label(), ctx)
        ctx.append_basic_block(else_block)

        # convert "else"
        else_ret_val = None
        else_syms = symbols.copy()
        if len(ir.args) == 3:
            else_ret_val = _convert_ir_bb(
                ctx, ir.args[2], else_syms, variables, allocated_variables.copy()
            )
            if isinstance(else_ret_val, IRLiteral):
                assert isinstance(else_ret_val.value, int)  # help mypy
                else_ret_val = ctx.get_basic_block().append_instruction("store", else_ret_val)
        after_else_syms = else_syms.copy()
        else_block = ctx.get_basic_block()

        # convert "then"
        then_block = IRBasicBlock(ctx.get_next_label(), ctx)
        ctx.append_basic_block(then_block)

        then_ret_val = _convert_ir_bb(ctx, ir.args[1], symbols, variables, allocated_variables)
        if isinstance(then_ret_val, IRLiteral):
            then_ret_val = ctx.get_basic_block().append_instruction("store", then_ret_val)

        current_bb.append_instruction("jnz", cont_ret, then_block.label, else_block.label)

        after_then_syms = symbols.copy()
        then_block = ctx.get_basic_block()

        # exit bb
        exit_label = ctx.get_next_label()
        exit_bb = IRBasicBlock(exit_label, ctx)
        exit_bb = ctx.append_basic_block(exit_bb)

        if_ret = None
        if then_ret_val is not None and else_ret_val is not None:
            if_ret = exit_bb.append_instruction(
                "phi", then_block.label, then_ret_val, else_block.label, else_ret_val
            )

        common_symbols = _get_symbols_common(after_then_syms, after_else_syms)
        for sym, val in common_symbols.items():
            ret = exit_bb.append_instruction(
                "phi", then_block.label, val[0], else_block.label, val[1]
            )
            old_var = symbols.get(sym, None)
            symbols[sym] = ret
            if old_var is not None:
                for idx, var_rec in allocated_variables.items():  # type: ignore
                    if var_rec.value == old_var.value:
                        allocated_variables[idx] = ret  # type: ignore

        if not else_block.is_terminated:
            else_block.append_instruction("jmp", exit_bb.label)

        if not then_block.is_terminated:
            then_block.append_instruction("jmp", exit_bb.label)

        return if_ret

    elif ir.value == "with":
        ret = _convert_ir_bb(
            ctx, ir.args[1], symbols, variables, allocated_variables
        )  # initialization

        # Handle with nesting with same symbol
        with_symbols = symbols.copy()

        sym = ir.args[0]
        if isinstance(ret, IRLiteral):
            new_var = ctx.get_basic_block().append_instruction("store", ret)  # type: ignore
            with_symbols[sym.value] = new_var
        else:
            with_symbols[sym.value] = ret  # type: ignore

        return _convert_ir_bb(ctx, ir.args[2], with_symbols, variables, allocated_variables)  # body
    elif ir.value == "goto":
        _append_jmp(ctx, IRLabel(ir.args[0].value))
    elif ir.value == "djump":
        args = [_convert_ir_bb(ctx, ir.args[0], symbols, variables, allocated_variables)]
        for target in ir.args[1:]:
            args.append(IRLabel(target.value))
        ctx.get_basic_block().append_instruction("djmp", *args)
        _new_block(ctx)
    elif ir.value == "set":
        sym = ir.args[0]
        arg_1 = _convert_ir_bb(ctx, ir.args[1], symbols, variables, allocated_variables)
        new_var = ctx.get_basic_block().append_instruction("store", arg_1)  # type: ignore
        symbols[sym.value] = new_var

    elif ir.value == "calldatacopy":
        arg_0, arg_1, size = _convert_ir_bb_list(
            ctx, ir.args, symbols, variables, allocated_variables
        )

        new_v = arg_0
        var = (
            _get_variable_from_address(variables, int(arg_0.value))
            if isinstance(arg_0, IRLiteral)
            else None
        )
        bb = ctx.get_basic_block()
        if var is not None:
            if allocated_variables.get(var.name, None) is None:
                new_v = bb.append_instruction("alloca", var.size, var.pos)  # type: ignore
                allocated_variables[var.name] = new_v  # type: ignore
            bb.append_instruction("calldatacopy", size, arg_1, new_v)  # type: ignore
            symbols[f"&{var.pos}"] = new_v  # type: ignore
        else:
            bb.append_instruction("calldatacopy", size, arg_1, new_v)  # type: ignore

        return new_v
    elif ir.value == "codecopy":
        arg_0, arg_1, size = _convert_ir_bb_list(
            ctx, ir.args, symbols, variables, allocated_variables
        )

        ctx.get_basic_block().append_instruction("codecopy", size, arg_1, arg_0)  # type: ignore
    elif ir.value == "symbol":
        return IRLabel(ir.args[0].value, True)
    elif ir.value == "data":
        label = IRLabel(ir.args[0].value)
        ctx.append_data("dbname", [label])
        for c in ir.args[1:]:
            if isinstance(c, int):
                assert 0 <= c <= 255, "data with invalid size"
                ctx.append_data("db", [c])  # type: ignore
            elif isinstance(c, bytes):
                ctx.append_data("db", [c])  # type: ignore
            elif isinstance(c, IRnode):
                data = _convert_ir_bb(ctx, c, symbols, variables, allocated_variables)
                ctx.append_data("db", [data])  # type: ignore
    elif ir.value == "assert":
        arg_0 = _convert_ir_bb(ctx, ir.args[0], symbols, variables, allocated_variables)
        current_bb = ctx.get_basic_block()
        current_bb.append_instruction("assert", arg_0)
    elif ir.value == "label":
        label = IRLabel(ir.args[0].value, True)
        bb = ctx.get_basic_block()
        if not bb.is_terminated:
            bb.append_instruction("jmp", label)
        bb = IRBasicBlock(label, ctx)
        ctx.append_basic_block(bb)
        _convert_ir_bb(ctx, ir.args[2], symbols, variables, allocated_variables)
    elif ir.value == "exit_to":
        func_t = ir.passthrough_metadata.get("func_t", None)
        assert func_t is not None, "exit_to without func_t"

        if func_t.is_external:
            # Hardcoded constructor special case
            bb = ctx.get_basic_block()
            if func_t.name == "__init__":
                label = IRLabel(ir.args[0].value, True)
                bb.append_instruction("jmp", label)
                return None
            if func_t.return_type is None:
                bb.append_instruction("stop")
                return None
            else:
                last_ir = None
                ret_var = ir.args[1]
                deleted = None
                if ret_var.is_literal and symbols.get(f"&{ret_var.value}", None) is not None:
                    deleted = symbols[f"&{ret_var.value}"]
                    del symbols[f"&{ret_var.value}"]
                for arg in ir.args[2:]:
                    last_ir = _convert_ir_bb(ctx, arg, symbols, variables, allocated_variables)
                if deleted is not None:
                    symbols[f"&{ret_var.value}"] = deleted

                ret_ir = _convert_ir_bb(ctx, ret_var, symbols, variables, allocated_variables)

                bb = ctx.get_basic_block()

                var = (
                    _get_variable_from_address(variables, int(ret_ir.value))
                    if isinstance(ret_ir, IRLiteral)
                    else None
                )
                if var is not None:
                    allocated_var = allocated_variables.get(var.name, None)
                    assert allocated_var is not None, "unallocated variable"
                    new_var = symbols.get(f"&{ret_ir.value}", allocated_var)  # type: ignore

                    if var.size and int(var.size) > 32:
                        offset = int(ret_ir.value) - var.pos  # type: ignore
                        if offset > 0:
                            ptr_var = bb.append_instruction("add", var.pos, offset)
                        else:
                            ptr_var = allocated_var
                        bb.append_instruction("return", last_ir, ptr_var)
                    else:
                        _append_return_for_stack_operand(ctx, symbols, new_var, last_ir)
                else:
                    if isinstance(ret_ir, IRLiteral):
                        sym = symbols.get(f"&{ret_ir.value}", None)
                        if sym is None:
                            bb.append_instruction("return", last_ir, ret_ir)
                        else:
                            if func_t.return_type.memory_bytes_required > 32:
                                new_var = bb.append_instruction("alloca", 32, ret_ir)
                                bb.append_instruction("mstore", sym, new_var)
                                bb.append_instruction("return", last_ir, new_var)
                            else:
                                bb.append_instruction("return", last_ir, ret_ir)
                    else:
                        if last_ir and int(last_ir.value) > 32:
                            bb.append_instruction("return", last_ir, ret_ir)
                        else:
                            ret_buf = 128  # TODO: need allocator
                            new_var = bb.append_instruction("alloca", 32, ret_buf)
                            bb.append_instruction("mstore", ret_ir, new_var)
                            bb.append_instruction("return", last_ir, new_var)

                ctx.append_basic_block(IRBasicBlock(ctx.get_next_label(), ctx))

        bb = ctx.get_basic_block()
        if func_t.is_internal:
            assert ir.args[1].value == "return_pc", "return_pc not found"
            if func_t.return_type is None:
                bb.append_instruction("ret", symbols["return_pc"])
            else:
                if func_t.return_type.memory_bytes_required > 32:
                    bb.append_instruction("ret", symbols["return_buffer"], symbols["return_pc"])
                else:
                    ret_by_value = bb.append_instruction("mload", symbols["return_buffer"])
                    bb.append_instruction("ret", ret_by_value, symbols["return_pc"])

    elif ir.value == "revert":
        arg_0, arg_1 = _convert_ir_bb_list(ctx, ir.args, symbols, variables, allocated_variables)
        ctx.get_basic_block().append_instruction("revert", arg_1, arg_0)

    elif ir.value == "dload":
        arg_0 = _convert_ir_bb(ctx, ir.args[0], symbols, variables, allocated_variables)
        bb = ctx.get_basic_block()
        src = bb.append_instruction("add", arg_0, IRLabel("code_end"))

        bb.append_instruction("dloadbytes", 32, src, MemoryPositions.FREE_VAR_SPACE)
        return bb.append_instruction("mload", MemoryPositions.FREE_VAR_SPACE)

    elif ir.value == "dloadbytes":
        dst, src_offset, len_ = _convert_ir_bb_list(
            ctx, ir.args, symbols, variables, allocated_variables
        )

        bb = ctx.get_basic_block()
        src = bb.append_instruction("add", src_offset, IRLabel("code_end"))
        bb.append_instruction("dloadbytes", len_, src, dst)
        return None

    elif ir.value == "mload":
        sym_ir = ir.args[0]
        var = (
            _get_variable_from_address(variables, int(sym_ir.value)) if sym_ir.is_literal else None
        )
        bb = ctx.get_basic_block()
        if var is not None:
            if var.size and var.size > 32:
                if allocated_variables.get(var.name, None) is None:
                    allocated_variables[var.name] = bb.append_instruction(
                        "alloca", var.size, var.pos
                    )

                offset = int(sym_ir.value) - var.pos
                if offset > 0:
                    ptr_var = bb.append_instruction("add", var.pos, offset)
                else:
                    ptr_var = allocated_variables[var.name]

                return bb.append_instruction("mload", ptr_var)
            else:
                if sym_ir.is_literal:
                    sym = symbols.get(f"&{sym_ir.value}", None)
                    if sym is None:
                        new_var = _convert_ir_bb(
                            ctx, sym_ir, symbols, variables, allocated_variables
                        )
                        symbols[f"&{sym_ir.value}"] = new_var
                        if allocated_variables.get(var.name, None) is None:
                            allocated_variables[var.name] = new_var
                            return new_var
                    else:
                        return sym

                sym = symbols.get(f"&{sym_ir.value}", None)
                assert sym is not None, "unallocated variable"
                return sym
        else:
            if sym_ir.is_literal:
                new_var = symbols.get(f"&{sym_ir.value}", None)
                if new_var is not None:
                    return bb.append_instruction("mload", new_var)
                else:
                    return bb.append_instruction("mload", sym_ir.value)
            else:
                new_var = _convert_ir_bb(ctx, sym_ir, symbols, variables, allocated_variables)
                #
                # Old IR gets it's return value as a reference in the stack
                # New IR gets it's return value in stack in case of 32 bytes or less
                # So here we detect ahead of time if this mload leads a self call and
                # and we skip the mload
                #
                if sym_ir.is_self_call:
                    return new_var
                return bb.append_instruction("mload", new_var)

    elif ir.value == "mstore":
        sym_ir, arg_1 = _convert_ir_bb_list(ctx, ir.args, symbols, variables, allocated_variables)

        bb = ctx.get_basic_block()

        var = None
        if isinstance(sym_ir, IRLiteral):
            var = _get_variable_from_address(variables, int(sym_ir.value))

        if var is not None and var.size is not None:
            if var.size and var.size > 32:
                if allocated_variables.get(var.name, None) is None:
                    allocated_variables[var.name] = bb.append_instruction(
                        "alloca", var.size, var.pos
                    )

                offset = int(sym_ir.value) - var.pos
                if offset > 0:
                    ptr_var = bb.append_instruction("add", var.pos, offset)
                else:
                    ptr_var = allocated_variables[var.name]

                bb.append_instruction("mstore", arg_1, ptr_var)
            else:
                if isinstance(sym_ir, IRLiteral):
                    new_var = bb.append_instruction("store", arg_1)
                    symbols[f"&{sym_ir.value}"] = new_var
                    # if allocated_variables.get(var.name, None) is None:
                    allocated_variables[var.name] = new_var
                return new_var
        else:
            if not isinstance(sym_ir, IRLiteral):
                bb.append_instruction("mstore", arg_1, sym_ir)
                return None

            sym = symbols.get(f"&{sym_ir.value}", None)
            if sym is None:
                bb.append_instruction("mstore", arg_1, sym_ir)
                if arg_1 and not isinstance(sym_ir, IRLiteral):
                    symbols[f"&{sym_ir.value}"] = arg_1
                return None

            if isinstance(sym_ir, IRLiteral):
                bb.append_instruction("mstore", arg_1, sym)
                return None
            else:
                symbols[sym_ir.value] = arg_1
                return arg_1
    elif ir.value == "ceil32":
        x = ir.args[0]
        expanded = IRnode.from_list(["and", ["add", x, 31], ["not", 31]])
        return _convert_ir_bb(ctx, expanded, symbols, variables, allocated_variables)
    elif ir.value == "select":
        # b ^ ((a ^ b) * cond) where cond is 1 or 0
        cond, a, b = ir.args
        expanded = IRnode.from_list(["xor", b, ["mul", cond, ["xor", a, b]]])
        return _convert_ir_bb(ctx, expanded, symbols, variables, allocated_variables)

    elif ir.value in ["sload", "iload"]:
        arg_0 = _convert_ir_bb(ctx, ir.args[0], symbols, variables, allocated_variables)
        return ctx.get_basic_block().append_instruction(ir.value, arg_0)
    elif ir.value in ["sstore", "istore"]:
        arg_0, arg_1 = _convert_ir_bb_list(ctx, ir.args, symbols, variables, allocated_variables)
        ctx.get_basic_block().append_instruction(ir.value, arg_1, arg_0)
    elif ir.value == "unique_symbol":
        sym = ir.args[0]
        new_var = ctx.get_next_variable()
        symbols[f"&{sym.value}"] = new_var
        return new_var
    elif ir.value == "repeat":
        #
        # repeat(sym, start, end, bound, body)
        # 1) entry block         ]
        # 2) init counter block  ] -> same block
        # 3) condition block (exit block, body block)
        # 4) body block
        # 5) increment block
        # 6) exit block
        # TODO: Add the extra bounds check after clarify
        def emit_body_blocks():
            global _break_target, _continue_target
            old_targets = _break_target, _continue_target
            _break_target, _continue_target = exit_block, increment_block
            _convert_ir_bb(ctx, body, symbols, variables, allocated_variables)
            _break_target, _continue_target = old_targets

        sym = ir.args[0]
        start, end, _ = _convert_ir_bb_list(
            ctx, ir.args[1:4], symbols, variables, allocated_variables
        )

        body = ir.args[4]

        entry_block = ctx.get_basic_block()
        cond_block = IRBasicBlock(ctx.get_next_label(), ctx)
        body_block = IRBasicBlock(ctx.get_next_label(), ctx)
        jump_up_block = IRBasicBlock(ctx.get_next_label(), ctx)
        increment_block = IRBasicBlock(ctx.get_next_label(), ctx)
        exit_block = IRBasicBlock(ctx.get_next_label(), ctx)

        counter_inc_var = ctx.get_next_variable()

        counter_var = ctx.get_basic_block().append_instruction("store", start)
        symbols[sym.value] = counter_var
        ctx.get_basic_block().append_instruction("jmp", cond_block.label)

        ret = cond_block.append_instruction(
            "phi", entry_block.label, counter_var, increment_block.label, counter_inc_var
        )
        symbols[sym.value] = ret

        xor_ret = cond_block.append_instruction("xor", ret, end)
        cont_ret = cond_block.append_instruction("iszero", xor_ret)
        ctx.append_basic_block(cond_block)

        start_syms = symbols.copy()
        ctx.append_basic_block(body_block)
        emit_body_blocks()
        end_syms = symbols.copy()
        diff_syms = _get_symbols_common(start_syms, end_syms)

        replacements = {}
        for sym, val in diff_syms.items():
            new_var = ctx.get_next_variable()
            symbols[sym] = new_var
            replacements[val[0]] = new_var
            replacements[val[1]] = new_var
            cond_block.insert_instruction(
                IRInstruction(
                    "phi", [entry_block.label, val[0], increment_block.label, val[1]], new_var
                ),
                1,
            )

        body_block.replace_operands(replacements)

        body_end = ctx.get_basic_block()
        if not body_end.is_terminated:
            body_end.append_instruction("jmp", jump_up_block.label)

        jump_up_block.append_instruction("jmp", increment_block.label)
        ctx.append_basic_block(jump_up_block)

        increment_block.insert_instruction(
            IRInstruction("add", [ret, IRLiteral(1)], counter_inc_var), 0
        )

        increment_block.append_instruction("jmp", cond_block.label)
        ctx.append_basic_block(increment_block)

        ctx.append_basic_block(exit_block)

        cond_block.append_instruction("jnz", cont_ret, exit_block.label, body_block.label)
    elif ir.value == "break":
        assert _break_target is not None, "Break with no break target"
        ctx.get_basic_block().append_instruction("jmp", _break_target.label)
        ctx.append_basic_block(IRBasicBlock(ctx.get_next_label(), ctx))
    elif ir.value == "continue":
        assert _continue_target is not None, "Continue with no contrinue target"
        ctx.get_basic_block().append_instruction("jmp", _continue_target.label)
        ctx.append_basic_block(IRBasicBlock(ctx.get_next_label(), ctx))
    elif ir.value == "gas":
        return ctx.get_basic_block().append_instruction("gas")
    elif ir.value == "returndatasize":
        return ctx.get_basic_block().append_instruction("returndatasize")
    elif ir.value == "returndatacopy":
        assert len(ir.args) == 3, "returndatacopy with wrong number of arguments"
        arg_0, arg_1, size = _convert_ir_bb_list(
            ctx, ir.args, symbols, variables, allocated_variables
        )

        new_var = ctx.get_basic_block().append_instruction("returndatacopy", arg_1, size)

        symbols[f"&{arg_0.value}"] = new_var
        return new_var
    elif ir.value == "selfdestruct":
        arg_0 = _convert_ir_bb(ctx, ir.args[0], symbols, variables, allocated_variables)
        ctx.get_basic_block().append_instruction("selfdestruct", arg_0)
    elif isinstance(ir.value, str) and ir.value.startswith("log"):
        args = reversed(
            [_convert_ir_bb(ctx, arg, symbols, variables, allocated_variables) for arg in ir.args]
        )
        topic_count = int(ir.value[3:])
        assert topic_count >= 0 and topic_count <= 4, "invalid topic count"
        ctx.get_basic_block().append_instruction("log", topic_count, *args)
    elif isinstance(ir.value, str) and ir.value.upper() in get_opcodes():
        _convert_ir_opcode(ctx, ir, symbols, variables, allocated_variables)
    elif isinstance(ir.value, str) and ir.value in symbols:
        return symbols[ir.value]
    elif ir.is_literal:
        return IRLiteral(ir.value)
    else:
        raise Exception(f"Unknown IR node: {ir}")

    return None


def _convert_ir_opcode(
    ctx: IRFunction,
    ir: IRnode,
    symbols: SymbolTable,
    variables: OrderedSet,
    allocated_variables: dict[str, IRVariable],
) -> None:
    opcode = ir.value.upper()  # type: ignore
    inst_args = []
    for arg in ir.args:
        if isinstance(arg, IRnode):
            inst_args.append(_convert_ir_bb(ctx, arg, symbols, variables, allocated_variables))
    ctx.get_basic_block().append_instruction(opcode, *inst_args)


def _data_ofst_of(sym, ofst, height_):
    # e.g. _OFST _sym_foo 32
    assert is_symbol(sym) or is_mem_sym(sym)
    if isinstance(ofst.value, int):
        # resolve at compile time using magic _OFST op
        return ["_OFST", sym, ofst.value]
    else:
        # if we can't resolve at compile time, resolve at runtime
        # ofst = _compile_to_assembly(ofst, withargs, existing_labels, break_dest, height_)
        return ofst + [sym, "ADD"]



from vyper.exceptions import FunctionDeclarationException, StateAccessViolation


def test_pure_operation(get_contract_with_gas_estimation_for_constants):
    c = get_contract_with_gas_estimation_for_constants(
        """
@pure
@external
def foo() -> int128:
    return 5
    """
    )
    assert c.foo() == 5


def test_pure_call(get_contract_with_gas_estimation_for_constants):
    c = get_contract_with_gas_estimation_for_constants(
        """
@pure
@internal
def _foo() -> int128:
    return 5

@pure
@external
def foo() -> int128:
    return self._foo()
    """
    )
    assert c.foo() == 5


def test_pure_interface(get_contract_with_gas_estimation_for_constants):
    c1 = get_contract_with_gas_estimation_for_constants(
        """
@pure
@external
def foo() -> int128:
    return 5
    """
    )
    c2 = get_contract_with_gas_estimation_for_constants(
        """
interface Foo:
    def foo() -> int128: pure

@pure
@external
def foo(a: address) -> int128:
    return staticcall Foo(a).foo()
    """
    )
    assert c2.foo(c1.address) == 5


def test_invalid_envar_access(get_contract, assert_compile_failed):
    assert_compile_failed(
        lambda: get_contract(
            """
@pure
@external
def foo() -> uint256:
    return chain.id
    """
        ),
        StateAccessViolation,
    )


def test_invalid_state_access(get_contract, assert_compile_failed):
    assert_compile_failed(
        lambda: get_contract(
            """
x: uint256

@pure
@external
def foo() -> uint256:
    return self.x
    """
        ),
        StateAccessViolation,
    )


def test_invalid_self_access(get_contract, assert_compile_failed):
    assert_compile_failed(
        lambda: get_contract(
            """
@pure
@external
def foo() -> address:
    return self
    """
        ),
        StateAccessViolation,
    )


def test_invalid_call(get_contract, assert_compile_failed):
    assert_compile_failed(
        lambda: get_contract(
            """
@view
@internal
def _foo() -> uint256:
    return 5

@pure
@external
def foo() -> uint256:
    return self._foo()  # Fails because of calling non-pure fn
    """
        ),
        StateAccessViolation,
    )


def test_invalid_conflicting_decorators(get_contract, assert_compile_failed):
    assert_compile_failed(
        lambda: get_contract(
            """
@pure
@external
@payable
def foo() -> uint256:
    return 5
    """
        ),
        FunctionDeclarationException,
    )



import json

from vyper import compiler
from vyper.ast.nodes import NODE_SRC_ATTRIBUTES
from vyper.ast.parse import parse_to_ast
from vyper.ast.utils import ast_to_dict, dict_to_ast


def get_node_ids(ast_struct, ids=None):
    if ids is None:
        ids = []

    for k, v in ast_struct.items():
        if isinstance(v, dict):
            ids = get_node_ids(v, ids)
        elif isinstance(v, list):
            for x in v:
                ids = get_node_ids(x, ids)
        elif k == "node_id":
            ids.append(v)
        elif v is None or isinstance(v, (str, int)):
            continue
        else:
            raise Exception(f"Unknown ast_struct provided. {k}, {v}")
    return ids


def test_ast_to_dict_node_id():
    code = """
@external
def test() -> int128:
    a: uint256 = 100
    return 123
    """
    dict_out = compiler.compile_code(code, output_formats=["ast_dict"])
    node_ids = get_node_ids(dict_out)

    assert len(node_ids) == len(set(node_ids))


def test_basic_ast():
    code = """
a: int128
    """
    dict_out = compiler.compile_code(code, output_formats=["annotated_ast_dict"], source_id=0)
    assert dict_out["annotated_ast_dict"]["ast"]["body"][0] == {
        "annotation": {
            "ast_type": "Name",
            "col_offset": 3,
            "end_col_offset": 9,
            "end_lineno": 2,
            "id": "int128",
            "lineno": 2,
            "node_id": 4,
            "src": "4:6:0",
        },
        "ast_type": "VariableDecl",
        "col_offset": 0,
        "end_col_offset": 9,
        "end_lineno": 2,
        "lineno": 2,
        "node_id": 1,
        "src": "1:9:0",
        "target": {
            "ast_type": "Name",
            "col_offset": 0,
            "end_col_offset": 1,
            "end_lineno": 2,
            "id": "a",
            "lineno": 2,
            "node_id": 2,
            "src": "1:1:0",
            "type": "int128",
        },
        "value": None,
        "is_constant": False,
        "is_immutable": False,
        "is_public": False,
        "is_transient": False,
        "type": "int128",
    }


def test_implements_ast():
    code = """
interface Foo:
    def foo() -> uint256: view

implements: Foo

@external
@view
def foo() -> uint256:
    return 1
    """
    dict_out = compiler.compile_code(code, output_formats=["ast_dict"], source_id=0)
    assert dict_out["ast_dict"]["ast"]["body"][1] == {
        "col_offset": 0,
        "annotation": {
            "col_offset": 12,
            "end_col_offset": 15,
            "node_id": 12,
            "src": "60:3:0",
            "ast_type": "Name",
            "end_lineno": 5,
            "lineno": 5,
            "id": "Foo",
        },
        "end_col_offset": 15,
        "node_id": 9,
        "src": "48:15:0",
        "ast_type": "ImplementsDecl",
        "end_lineno": 5,
        "lineno": 5,
    }


def test_dict_to_ast():
    code = """
@external
def test() -> int128:
    a: uint256 = 100
    b: int128 = -22
    c: decimal = -3.3133700
    d: Bytes[11] = b"oh hai mark"
    e: Bytes[1] = 0b01010101
    f: Bytes[88] = b"\x01\x70"
    g: String[100] = "  baka baka   "
    h: address = 0xdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef
    i: bool = False
    return 123
    """

    original_ast = parse_to_ast(code)
    out_dict = ast_to_dict(original_ast)
    out_json = json.dumps(out_dict)
    new_dict = json.loads(out_json)
    new_ast = dict_to_ast(new_dict)

    assert new_ast == original_ast


# strip source annotations like lineno, we don't care for inspecting
# the analysis result
def _strip_source_annotations(dict_node):
    to_strip = NODE_SRC_ATTRIBUTES + ("node_id",)
    if isinstance(dict_node, dict):
        for k in list(dict_node.keys()):
            if k in to_strip:
                del dict_node[k]
                continue
            _strip_source_annotations(dict_node[k])
    elif isinstance(dict_node, list):
        for child in dict_node:
            _strip_source_annotations(child)


def test_output_variable_read_write_analysis(make_input_bundle, chdir_tmp_path):
    # test we output the result of variable read/write correctly
    lib1 = """
struct Foo:
    a: uint256
    b: String[3]
    c: decimal

struct Bar:
    items: Foo[2]

counter: uint256

bars: DynArray[Bar, 10]
    """
    code = """
import lib1

initializes: lib1

counter: uint256

@internal
def foo():
    x: uint256 = lib1.counter
    lib1.counter += 1

@internal
def bar():
    x: uint256 = lib1.counter
    y: uint256 = self.counter
    lib1.counter += 1

@internal
def baz():
    self.bar()  # reads both lib1.counter and self.counter
    self.counter += 1

@internal
def qux():
    lib1.bars = []
    lib1.bars[0] = empty(lib1.Bar)
    lib1.bars[1].items = empty(lib1.Foo[2])

    lib1.bars[1].items[0].a = 1
    lib1.bars[0].items[1].c = 10.0

@internal
def qux2():
    self.qux()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})

    out = compiler.compile_code(
        code,
        contract_path="main.vy",
        input_bundle=input_bundle,
        output_formats=["annotated_ast_dict"],
        source_id=0,
    )["annotated_ast_dict"]["ast"]
    _strip_source_annotations(out)

    foo, bar, baz, qux, qux2 = out["body"][3:]
    assert foo["name"] == "foo"
    assert foo["body"] == [
        {
            "annotation": {"ast_type": "Name", "id": "uint256"},
            "ast_type": "AnnAssign",
            "target": {"ast_type": "Name", "id": "x", "type": "uint256"},
            "value": {
                "ast_type": "Attribute",
                "attr": "counter",
                "type": "uint256",
                "value": {"ast_type": "Name", "id": "lib1", "type": "lib1.vy"},
                "variable_reads": [{"access_path": [], "module": "lib1.vy", "variable": "counter"}],
            },
        },
        {
            "ast_type": "AugAssign",
            "op": {"ast_type": "Add"},
            "target": {
                "ast_type": "Attribute",
                "attr": "counter",
                "type": "uint256",
                "value": {"ast_type": "Name", "id": "lib1", "type": "lib1.vy"},
                "variable_reads": [{"access_path": [], "module": "lib1.vy", "variable": "counter"}],
                "variable_writes": [
                    {"access_path": [], "module": "lib1.vy", "variable": "counter"}
                ],
            },
            "value": {"ast_type": "Int", "type": "uint256", "value": 1},
        },
    ]

    assert bar["name"] == "bar"
    assert bar["body"] == [
        {
            "annotation": {"ast_type": "Name", "id": "uint256"},
            "ast_type": "AnnAssign",
            "target": {"ast_type": "Name", "id": "x", "type": "uint256"},
            "value": {
                "ast_type": "Attribute",
                "attr": "counter",
                "type": "uint256",
                "value": {"ast_type": "Name", "id": "lib1", "type": "lib1.vy"},
                "variable_reads": [{"access_path": [], "module": "lib1.vy", "variable": "counter"}],
            },
        },
        {
            "annotation": {"ast_type": "Name", "id": "uint256"},
            "ast_type": "AnnAssign",
            "target": {"ast_type": "Name", "id": "y", "type": "uint256"},
            "value": {
                "ast_type": "Attribute",
                "attr": "counter",
                "type": "uint256",
                "value": {"ast_type": "Name", "id": "self", "type": "self"},
                "variable_reads": [{"access_path": [], "module": "main.vy", "variable": "counter"}],
            },
        },
        {
            "ast_type": "AugAssign",
            "op": {"ast_type": "Add"},
            "target": {
                "ast_type": "Attribute",
                "attr": "counter",
                "type": "uint256",
                "value": {"ast_type": "Name", "id": "lib1", "type": "lib1.vy"},
                "variable_reads": [{"access_path": [], "module": "lib1.vy", "variable": "counter"}],
                "variable_writes": [
                    {"access_path": [], "module": "lib1.vy", "variable": "counter"}
                ],
            },
            "value": {"ast_type": "Int", "type": "uint256", "value": 1},
        },
    ]

    assert baz["name"] == "baz"
    assert baz["body"] == [
        {
            "ast_type": "Expr",
            "value": {
                "args": [],
                "ast_type": "Call",
                "func": {
                    "ast_type": "Attribute",
                    "attr": "bar",
                    "type": "def bar():",
                    "value": {"ast_type": "Name", "id": "self", "type": "self"},
                    "variable_reads": [
                        {"access_path": [], "module": "lib1.vy", "variable": "counter"},
                        {"access_path": [], "module": "main.vy", "variable": "counter"},
                    ],
                    "variable_writes": [
                        {"access_path": [], "module": "lib1.vy", "variable": "counter"}
                    ],
                },
                "keywords": [],
                "type": "(void)",
            },
        },
        {
            "ast_type": "AugAssign",
            "op": {"ast_type": "Add"},
            "target": {
                "ast_type": "Attribute",
                "attr": "counter",
                "type": "uint256",
                "value": {"ast_type": "Name", "id": "self", "type": "self"},
                "variable_reads": [{"access_path": [], "module": "main.vy", "variable": "counter"}],
                "variable_writes": [
                    {"access_path": [], "module": "main.vy", "variable": "counter"}
                ],
            },
            "value": {"ast_type": "Int", "type": "uint256", "value": 1},
        },
    ]

    assert qux["name"] == "qux"
    assert qux["body"] == [
        {
            "ast_type": "Assign",
            "target": {
                "ast_type": "Attribute",
                "attr": "bars",
                "type": "DynArray[Bar declaration object, 10]",
                "value": {"ast_type": "Name", "id": "lib1", "type": "lib1.vy"},
                "variable_reads": [{"access_path": [], "module": "lib1.vy", "variable": "bars"}],
                "variable_writes": [{"access_path": [], "module": "lib1.vy", "variable": "bars"}],
            },
            "value": {
                "ast_type": "List",
                "elements": [],
                "type": "DynArray[Bar declaration object, 10]",
            },
        },
        {
            "ast_type": "Assign",
            "target": {
                "ast_type": "Subscript",
                "slice": {"ast_type": "Int", "type": "int8", "value": 0},
                "type": "Bar declaration object",
                "value": {
                    "ast_type": "Attribute",
                    "attr": "bars",
                    "type": "DynArray[Bar declaration object, 10]",
                    "value": {"ast_type": "Name", "id": "lib1", "type": "lib1.vy"},
                    "variable_reads": [
                        {"access_path": [], "module": "lib1.vy", "variable": "bars"}
                    ],
                },
                "variable_reads": [
                    {"access_path": ["$subscript_access"], "module": "lib1.vy", "variable": "bars"}
                ],
                "variable_writes": [
                    {"access_path": ["$subscript_access"], "module": "lib1.vy", "variable": "bars"}
                ],
            },
            "value": {
                "args": [
                    {
                        "ast_type": "Attribute",
                        "attr": "Bar",
                        "type": "type(Bar declaration object)",
                        "value": {"ast_type": "Name", "id": "lib1", "type": "lib1.vy"},
                    }
                ],
                "ast_type": "Call",
                "func": {"ast_type": "Name", "id": "empty", "type": "(builtin) empty"},
                "keywords": [],
                "type": "Bar declaration object",
            },
        },
        {
            "ast_type": "Assign",
            "target": {
                "ast_type": "Attribute",
                "attr": "items",
                "type": "Foo declaration object[2]",
                "value": {
                    "ast_type": "Subscript",
                    "slice": {"ast_type": "Int", "type": "int8", "value": 1},
                    "type": "Bar declaration object",
                    "value": {
                        "ast_type": "Attribute",
                        "attr": "bars",
                        "type": "DynArray[Bar declaration object, 10]",
                        "value": {"ast_type": "Name", "id": "lib1", "type": "lib1.vy"},
                        "variable_reads": [
                            {"access_path": [], "module": "lib1.vy", "variable": "bars"}
                        ],
                    },
                    "variable_reads": [
                        {
                            "access_path": ["$subscript_access"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        }
                    ],
                },
                "variable_reads": [
                    {
                        "access_path": ["$subscript_access", "items"],
                        "module": "lib1.vy",
                        "variable": "bars",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": ["$subscript_access", "items"],
                        "module": "lib1.vy",
                        "variable": "bars",
                    }
                ],
            },
            "value": {
                "args": [
                    {
                        "ast_type": "Subscript",
                        "slice": {"ast_type": "Int", "value": 2},
                        "type": "type(Foo declaration object[2])",
                        "value": {
                            "ast_type": "Attribute",
                            "attr": "Foo",
                            "value": {"ast_type": "Name", "id": "lib1"},
                        },
                    }
                ],
                "ast_type": "Call",
                "func": {"ast_type": "Name", "id": "empty", "type": "(builtin) empty"},
                "keywords": [],
                "type": "Foo declaration object[2]",
            },
        },
        {
            "ast_type": "Assign",
            "target": {
                "ast_type": "Attribute",
                "attr": "a",
                "type": "uint256",
                "value": {
                    "ast_type": "Subscript",
                    "slice": {"ast_type": "Int", "type": "int8", "value": 0},
                    "type": "Foo declaration object",
                    "value": {
                        "ast_type": "Attribute",
                        "attr": "items",
                        "type": "Foo declaration object[2]",
                        "value": {
                            "ast_type": "Subscript",
                            "slice": {"ast_type": "Int", "type": "int8", "value": 1},
                            "type": "Bar declaration object",
                            "value": {
                                "ast_type": "Attribute",
                                "attr": "bars",
                                "type": "DynArray[Bar " "declaration " "object, 10]",
                                "value": {"ast_type": "Name", "id": "lib1", "type": "lib1.vy"},
                                "variable_reads": [
                                    {"access_path": [], "module": "lib1.vy", "variable": "bars"}
                                ],
                            },
                            "variable_reads": [
                                {
                                    "access_path": ["$subscript_access"],
                                    "module": "lib1.vy",
                                    "variable": "bars",
                                }
                            ],
                        },
                        "variable_reads": [
                            {
                                "access_path": ["$subscript_access", "items"],
                                "module": "lib1.vy",
                                "variable": "bars",
                            }
                        ],
                    },
                    "variable_reads": [
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        }
                    ],
                },
                "variable_reads": [
                    {
                        "access_path": ["$subscript_access", "items", "$subscript_access", "a"],
                        "module": "lib1.vy",
                        "variable": "bars",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": ["$subscript_access", "items", "$subscript_access", "a"],
                        "module": "lib1.vy",
                        "variable": "bars",
                    }
                ],
            },
            "value": {"ast_type": "Int", "type": "uint256", "value": 1},
        },
        {
            "ast_type": "Assign",
            "target": {
                "ast_type": "Attribute",
                "attr": "c",
                "type": "decimal",
                "value": {
                    "ast_type": "Subscript",
                    "slice": {"ast_type": "Int", "type": "int8", "value": 1},
                    "type": "Foo declaration object",
                    "value": {
                        "ast_type": "Attribute",
                        "attr": "items",
                        "type": "Foo declaration object[2]",
                        "value": {
                            "ast_type": "Subscript",
                            "slice": {"ast_type": "Int", "type": "int8", "value": 0},
                            "type": "Bar declaration object",
                            "value": {
                                "ast_type": "Attribute",
                                "attr": "bars",
                                "type": "DynArray[Bar " "declaration " "object, 10]",
                                "value": {"ast_type": "Name", "id": "lib1", "type": "lib1.vy"},
                                "variable_reads": [
                                    {"access_path": [], "module": "lib1.vy", "variable": "bars"}
                                ],
                            },
                            "variable_reads": [
                                {
                                    "access_path": ["$subscript_access"],
                                    "module": "lib1.vy",
                                    "variable": "bars",
                                }
                            ],
                        },
                        "variable_reads": [
                            {
                                "access_path": ["$subscript_access", "items"],
                                "module": "lib1.vy",
                                "variable": "bars",
                            }
                        ],
                    },
                    "variable_reads": [
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        }
                    ],
                },
                "variable_reads": [
                    {
                        "access_path": ["$subscript_access", "items", "$subscript_access", "c"],
                        "module": "lib1.vy",
                        "variable": "bars",
                    }
                ],
                "variable_writes": [
                    {
                        "access_path": ["$subscript_access", "items", "$subscript_access", "c"],
                        "module": "lib1.vy",
                        "variable": "bars",
                    }
                ],
            },
            "value": {"ast_type": "Decimal", "type": "decimal", "value": "10.0"},
        },
    ]

    assert qux2["name"] == "qux2"
    assert qux2["body"] == [
        {
            "ast_type": "Expr",
            "value": {
                "args": [],
                "ast_type": "Call",
                "func": {
                    "ast_type": "Attribute",
                    "attr": "qux",
                    "type": "def qux():",
                    "value": {"ast_type": "Name", "id": "self", "type": "self"},
                    "variable_reads": [
                        {"access_path": [], "module": "lib1.vy", "variable": "bars"},
                        {
                            "access_path": ["$subscript_access"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access", "a"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access", "c"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        },
                    ],
                    "variable_writes": [
                        {"access_path": [], "module": "lib1.vy", "variable": "bars"},
                        {
                            "access_path": ["$subscript_access"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access", "a"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        },
                        {
                            "access_path": ["$subscript_access", "items", "$subscript_access", "c"],
                            "module": "lib1.vy",
                            "variable": "bars",
                        },
                    ],
                },
                "keywords": [],
                "type": "(void)",
            },
        }
    ]



import pytest


@pytest.fixture
def c(w3, get_contract):
    with open("examples/stock/company.vy") as f:
        contract_code = f.read()
        contract = get_contract(contract_code, *[w3.eth.accounts[0], 1000, 10**6])
    return contract


def test_overbuy(w3, c, tx_failed):
    # If all the stock has been bought, no one can buy more
    a1, a2 = w3.eth.accounts[1:3]
    test_shares = int(c.totalShares() / 2)
    test_value = int(test_shares * c.price())
    c.buyStock(transact={"from": a1, "value": test_value})
    c.buyStock(transact={"from": a1, "value": test_value})
    assert c.stockAvailable() == 0
    assert c.getHolding(a1) == (test_shares * 2)
    one_stock = c.price()
    with tx_failed():
        c.buyStock(transact={"from": a1, "value": one_stock})
    with tx_failed():
        c.buyStock(transact={"from": a2, "value": one_stock})


def test_sell_without_stock(w3, c, tx_failed):
    a1, a2 = w3.eth.accounts[1:3]
    # If you don't have any stock, you can't sell
    with tx_failed():
        c.sellStock(1, transact={"from": a1})
    with tx_failed():
        c.sellStock(1, transact={"from": a2})
    # But if you do, you can!
    test_shares = int(c.totalShares())
    test_value = int(test_shares * c.price())
    c.buyStock(transact={"from": a1, "value": test_value})
    assert c.getHolding(a1) == test_shares
    c.sellStock(test_shares, transact={"from": a1})
    # But only until you run out
    with tx_failed():
        c.sellStock(1, transact={"from": a1})


def test_oversell(w3, c, tx_failed):
    a0, a1, a2 = w3.eth.accounts[:3]
    # You can't sell more than you own
    test_shares = int(c.totalShares())
    test_value = int(test_shares * c.price())
    c.buyStock(transact={"from": a1, "value": test_value})
    with tx_failed():
        c.sellStock(test_shares + 1, transact={"from": a1})


def test_transfer(w3, c, tx_failed):
    # If you don't have any stock, you can't transfer
    a1, a2 = w3.eth.accounts[1:3]
    with tx_failed():
        c.transferStock(a2, 1, transact={"from": a1})
    with tx_failed():
        c.transferStock(a1, 1, transact={"from": a2})
    # If you transfer, you don't have the stock anymore
    test_shares = int(c.totalShares())
    test_value = int(test_shares * c.price())
    c.buyStock(transact={"from": a1, "value": test_value})
    assert c.getHolding(a1) == test_shares
    c.transferStock(a2, test_shares, transact={"from": a1})
    with tx_failed():
        c.sellStock(1, transact={"from": a1})
    # But the other person does
    c.sellStock(test_shares, transact={"from": a2})


def test_paybill(w3, c, tx_failed):
    a0, a1, a2, a3 = w3.eth.accounts[:4]
    # Only the company can authorize payments
    with tx_failed():
        c.payBill(a2, 1, transact={"from": a1})
    # A company can only pay someone if it has the money
    with tx_failed():
        c.payBill(a2, 1, transact={"from": a0})
    # If it has the money, it can pay someone
    test_value = int(c.totalShares() * c.price())
    c.buyStock(transact={"from": a1, "value": test_value})
    c.payBill(a2, test_value, transact={"from": a0})
    # Until it runs out of money
    with tx_failed():
        c.payBill(a3, 1, transact={"from": a0})
    # Then no stockholders can sell their stock either
    with tx_failed():
        c.sellStock(1, transact={"from": a1})


def test_valuation(w3, c):
    a1 = w3.eth.accounts[1]
    # Valuation is number of shares held times price
    assert c.debt() == 0
    test_value = int(c.totalShares() * c.price())
    c.buyStock(transact={"from": a1, "value": test_value})
    assert c.debt() == test_value


def test_logs(w3, c, get_logs):
    a0, a1, a2, a3 = w3.eth.accounts[:4]
    # Buy is logged
    logs = get_logs(c.buyStock(transact={"from": a1, "value": 7 * c.price()}), c, "Buy")
    assert len(logs) == 1
    assert logs[0].args.buy_order == 7

    # Sell is logged
    logs = get_logs(c.sellStock(3, transact={"from": a1}), c, "Sell")
    assert len(logs) == 1
    assert logs[0].args.sell_order == 3

    # Transfer is logged
    logs = get_logs(c.transferStock(a2, 4, transact={"from": a1}), c, "Transfer")
    assert len(logs) == 1
    assert logs[0].event == "Transfer"
    assert logs[0].args.value == 4

    # Pay is logged
    amount = 10**4
    logs = get_logs(c.payBill(a3, amount, transact={}), c, "Pay")
    assert len(logs) == 1
    assert logs[0].args.amount == amount



def test_block_number(get_contract_with_gas_estimation, w3):
    block_number_code = """
@external
def block_number() -> uint256:
    return block.number
    """
    c = get_contract_with_gas_estimation(block_number_code)

    assert c.block_number() == 1
    w3.testing.mine(1)
    assert c.block_number() == 2



from vyper.utils import OrderedSet
from vyper.venom.analysis import DFG
from vyper.venom.basicblock import IRBasicBlock, IRInstruction
from vyper.venom.function import IRFunction
from vyper.venom.passes.base_pass import IRPass


# DataFlow Transformation
class DFTPass(IRPass):
    def _process_instruction_r(self, bb: IRBasicBlock, inst: IRInstruction):
        if inst in self.visited_instructions:
            return
        self.visited_instructions.add(inst)

        if inst.opcode == "phi":
            # phi instructions stay at the beginning of the basic block
            # and no input processing is needed
            bb.instructions.append(inst)
            return

        for op in inst.get_inputs():
            target = self.dfg.get_producing_instruction(op)
            if target.parent != inst.parent or target.fence_id != inst.fence_id:
                # don't reorder across basic block or fence boundaries
                continue
            self._process_instruction_r(bb, target)

        bb.instructions.append(inst)

    def _process_basic_block(self, bb: IRBasicBlock) -> None:
        self.ctx.append_basic_block(bb)

        instructions = bb.instructions
        bb.instructions = []

        for inst in instructions:
            inst.fence_id = self.fence_id
            if inst.volatile:
                self.fence_id += 1

        for inst in instructions:
            self._process_instruction_r(bb, inst)

    def _run_pass(self, ctx: IRFunction) -> None:
        self.ctx = ctx
        self.dfg = DFG.build_dfg(ctx)
        self.fence_id = 0
        self.visited_instructions: OrderedSet[IRInstruction] = OrderedSet()

        basic_blocks = ctx.basic_blocks
        ctx.basic_blocks = []

        for bb in basic_blocks:
            self._process_basic_block(bb)



import re
import textwrap

import hypothesis
import hypothesis.strategies as st
import pytest
from hypothesis import HealthCheck, assume, given
from hypothesis.extra.lark import LarkStrategy

from vyper.ast import Module, parse_to_ast
from vyper.ast.grammar import parse_vyper_source, vyper_grammar
from vyper.ast.pre_parser import pre_parse


def test_basic_grammar():
    code = """
    a: uint256
    b: uint128
    """
    code_func = """
    @external
    def one_two_three() -> uint256:
        return 123123123
    """

    assert parse_vyper_source(code, dedent=True)
    assert parse_to_ast(textwrap.dedent(code))

    assert parse_vyper_source(code_func, dedent=True)
    assert parse_to_ast(textwrap.dedent(code_func))


def test_basic_grammar_empty():
    code = """
    """
    tree = parse_vyper_source(code, dedent=True)
    assert len(tree.children) == 0


def utf8_encodable(terminal: str) -> bool:
    try:
        if "\x00" not in terminal and "\\ " not in terminal and "\x0c" not in terminal:
            terminal.encode("utf-8-sig")
            return True
        else:
            return False
    except UnicodeEncodeError:  # pragma: no cover
        # Very rarely, a "." in some terminal regex will generate a surrogate
        # character that cannot be encoded as UTF-8.  We apply this filter to
        # ensure it doesn't happen at runtime, but don't worry about coverage.
        return False


# With help from hyposmith
# https://github.com/Zac-HD/hypothesmith/blob/master/src/hypothesmith/syntactic.py
class GrammarStrategy(LarkStrategy):
    def __init__(self, grammar, start, explicit_strategies):
        super().__init__(grammar, start, explicit_strategies)
        self.terminal_strategies = {
            k: v.map(lambda s: s.replace("\0", "")).filter(utf8_encodable)
            for k, v in self.terminal_strategies.items()  # type: ignore
        }

    def draw_symbol(self, data, symbol, draw_state):  # type: ignore
        count = len(draw_state.result)
        super().draw_symbol(data, symbol, draw_state)
        try:
            compile(
                source="".join(draw_state.result[count:])
                .replace("contract", "class")
                .replace("struct", "class"),  # HACK: Python ast.parse
                filename="<string>",
                mode="exec",
            )
        except SyntaxError:
            # Python's grammar doesn't actually fully describe the behaviour of the
            # CPython parser and AST-post-processor, so we just filter out errors.
            assume(False)


def from_grammar() -> st.SearchStrategy[str]:
    """
    Generate syntactically-valid Python source code based on the grammar.
    """
    grammar = vyper_grammar()
    explicit_strategies = dict(
        _INDENT=st.just(" " * 4),
        _DEDENT=st.just(""),
        NAME=st.from_regex(r"[a-z_A-Z]+", fullmatch=True).filter(str.isidentifier),
    )
    return GrammarStrategy(grammar, "module", explicit_strategies)


# Avoid examples with *only* single or double quote docstrings
# because they trigger a trivial parser bug
SINGLE_QUOTE_DOCSTRING = re.compile(r"^'''.*'''$")
DOUBLE_QUOTE_DOCSTRING = re.compile(r'^""".*"""$')


def has_no_docstrings(c):
    return not (SINGLE_QUOTE_DOCSTRING.match(c) or DOUBLE_QUOTE_DOCSTRING.match(c))


@pytest.mark.fuzzing
@given(code=from_grammar().filter(lambda c: utf8_encodable(c)))
@hypothesis.settings(max_examples=500, suppress_health_check=[HealthCheck.too_slow])
def test_grammar_bruteforce(code):
    if utf8_encodable(code):
        _, _, _, reformatted_code = pre_parse(code + "\n")
        tree = parse_to_ast(reformatted_code)
        assert isinstance(tree, Module)
